import asyncio
import enum
import json
import logging
from contextlib import asynccontextmanager, suppress
from dataclasses import dataclass, field
from datetime import datetime, timezone
from decimal import Decimal
from functools import partial
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    overload,
)

import aiohttp
import aiohttp.hdrs
import attr
from aiodocker.exceptions import DockerError
from aiohttp import WSMsgType, WSServerHandshakeError
from dateutil.parser import isoparse
from multidict import MultiDict
from yarl import URL

from ._abc import (
    AbstractDockerImageProgress,
    ImageCommitFinished,
    ImageCommitStarted,
    ImageProgressPush,
    ImageProgressSave,
)
from ._config import Config
from ._core import _Core
from ._errors import NDJSONError, ResourceNotFound, StdStreamError
from ._images import (
    _DummyProgress,
    _raise_on_error_chunk,
    _try_parse_image_progress_step,
)
from ._parser import DiskVolume, Parser, SecretFile, Volume
from ._parsing_utils import LocalImage, RemoteImage
from ._rewrite import rewrite_module
from ._url_utils import (
    normalize_disk_uri,
    normalize_secret_uri,
    normalize_storage_path_uri,
)
from ._utils import NoPublicConstructor, asyncgeneratorcontextmanager

log = logging.getLogger(__package__)

INVALID_IMAGE_NAME = "INVALID-IMAGE-NAME"


@rewrite_module
@dataclass(frozen=True)
class Resources:
    memory_mb: int
    cpu: float
    gpu: Optional[int] = None
    gpu_model: Optional[str] = None
    shm: bool = True
    tpu_type: Optional[str] = None
    tpu_software_version: Optional[str] = None


@rewrite_module
class JobStatus(str, enum.Enum):
    """An Enum subclass that represents job statuses.

    PENDING: a job is being created and scheduled. This includes finding (and
    possibly waiting for) sufficient amount of resources, pulling an image
    from a registry etc.
    SUSPENDED: a preemptible job is paused to allow other jobs to run.
    RUNNING: a job is being run.
    SUCCEEDED: a job terminated with the 0 exit code.
    CANCELLED: a running job was manually terminated/deleted.
    FAILED: a job terminated with a non-0 exit code.
    """

    PENDING = "pending"
    SUSPENDED = "suspended"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"
    UNKNOWN = "unknown"  # invalid status code, a default value is status is not sent

    @property
    def is_pending(self) -> bool:
        cls = type(self)
        return self in (cls.PENDING, cls.SUSPENDED)

    @property
    def is_running(self) -> bool:
        return self == type(self).RUNNING

    @property
    def is_finished(self) -> bool:
        cls = type(self)
        return self in (cls.SUCCEEDED, cls.FAILED, cls.CANCELLED)

    @classmethod
    def items(cls) -> Set["JobStatus"]:
        return {item for item in cls if item != cls.UNKNOWN}

    @classmethod
    def active_items(cls) -> Set["JobStatus"]:
        return {item for item in cls.items() if not item.is_finished}

    @classmethod
    def finished_items(cls) -> Set["JobStatus"]:
        return {item for item in cls.items() if item.is_finished}

    __format__ = str.__format__  # type: ignore[assignment]
    __str__ = str.__str__


@rewrite_module
@dataclass(frozen=True)
class HTTPPort:
    port: int
    requires_auth: bool = True


@rewrite_module
@dataclass(frozen=True)
class Container:
    image: RemoteImage
    resources: Resources
    entrypoint: Optional[str] = None
    command: Optional[str] = None
    working_dir: Optional[str] = None
    http: Optional[HTTPPort] = None
    env: Mapping[str, str] = field(default_factory=dict)
    volumes: Sequence[Volume] = field(default_factory=list)
    secret_env: Mapping[str, URL] = field(default_factory=dict)
    secret_files: Sequence[SecretFile] = field(default_factory=list)
    disk_volumes: Sequence[DiskVolume] = field(default_factory=list)
    tty: bool = False


@rewrite_module
@dataclass(frozen=True)
class JobStatusItem:
    status: JobStatus
    transition_time: datetime
    reason: str = ""
    description: str = ""
    exit_code: Optional[int] = None


@rewrite_module
@dataclass(frozen=True)
class JobStatusHistory:
    status: JobStatus
    reason: str
    description: str
    restarts: int = 0
    created_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    run_time_seconds: Optional[int] = None
    exit_code: Optional[int] = None
    transitions: Sequence[JobStatusItem] = field(default_factory=list)

    @property
    def changed_at(self) -> datetime:
        if self.status == JobStatus.PENDING:
            when = self.created_at
        elif self.status == JobStatus.RUNNING:
            when = self.started_at
        elif self.status == JobStatus.RUNNING or self.status == JobStatus.SUSPENDED:
            when = self.started_at
        elif self.status.is_finished:
            when = self.finished_at
        else:
            when = self.transitions[-1].transition_time
        assert when is not None
        return when


@rewrite_module
class JobRestartPolicy(str, enum.Enum):
    NEVER = "never"
    ON_FAILURE = "on-failure"
    ALWAYS = "always"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return repr(self.value)


@rewrite_module
@dataclass(frozen=True)
class JobDescriptionInternal:
    materialized: bool = False
    being_dropped: bool = False
    logs_removed: bool = False


@rewrite_module
@dataclass(frozen=True)
class JobDescription:
    id: str
    owner: str
    cluster_name: str
    status: JobStatus
    history: JobStatusHistory
    container: Container
    scheduler_enabled: bool
    pass_config: bool
    uri: URL
    total_price_credits: Decimal
    price_credits_per_hour: Decimal
    name: Optional[str] = None
    tags: Sequence[str] = ()
    description: Optional[str] = None
    http_url: URL = URL()
    internal_hostname: Optional[str] = None
    internal_hostname_named: Optional[str] = None
    restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER
    life_span: Optional[float] = None
    schedule_timeout: Optional[float] = None
    preset_name: Optional[str] = None
    org_name: Optional[str] = None
    preemptible_node: bool = False
    privileged: bool = False
    _internal: JobDescriptionInternal = JobDescriptionInternal()


@rewrite_module
@dataclass(frozen=True)
class JobTelemetry:
    cpu: float
    memory: float
    timestamp: float
    gpu_duty_cycle: Optional[int] = None
    gpu_memory: Optional[float] = None


@rewrite_module
@dataclass(frozen=True)
class Message:
    fileno: int
    data: bytes


@rewrite_module
class StdStream:
    def __init__(self, ws: aiohttp.ClientWebSocketResponse) -> None:
        self._ws = ws
        self._closing = False

    async def close(self) -> None:
        self._closing = True
        await self._ws.close()

    async def read_out(self) -> Optional[Message]:
        if self._closing:
            return None
        msg = await self._ws.receive()
        if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSING, WSMsgType.CLOSED):
            self._closing = True
            return None
        if msg.data[0] == 3:
            try:
                details = json.loads(msg.data[1:])
                exit_code = details["exit_code"]
            except Exception:
                exit_code = 1
            self._closing = True
            raise StdStreamError(exit_code)
        return Message(msg.data[0], msg.data[1:])

    async def write_in(self, data: bytes) -> None:
        if self._closing:
            return
        await self._ws.send_bytes(b"\x00" + data)

    async def resize(self, h: int, w: int) -> None:
        if self._closing:
            return
        await self._ws.send_bytes(b"\x04" + f'{{"height":{h},"width":{w}}}'.encode())


@rewrite_module
class Jobs(metaclass=NoPublicConstructor):
    def __init__(self, core: _Core, config: Config, parse: Parser) -> None:
        self._core = core
        self._config = config
        self._parse = parse

    def _get_monitoring_url(self, cluster_name: Optional[str]) -> URL:
        if cluster_name is None:
            cluster_name = self._config.cluster_name
        return self._config.get_cluster(cluster_name).monitoring_url

    async def run(
        self,
        container: Container,
        *,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
        description: Optional[str] = None,
        scheduler_enabled: bool = False,
        pass_config: bool = False,
        wait_for_jobs_quota: bool = False,
        schedule_timeout: Optional[float] = None,
        restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
        life_span: Optional[float] = None,
        org_name: Optional[str] = None,
    ) -> JobDescription:
        url = self._config.api_url / "jobs"
        payload = _job_to_api(
            cluster_name=self._config.cluster_name,
            name=name,
            tags=tags,
            description=description,
            pass_config=pass_config,
            wait_for_jobs_quota=wait_for_jobs_quota,
            schedule_timeout=schedule_timeout,
            restart_policy=restart_policy,
            life_span=life_span,
            org_name=org_name,
        )
        payload["container"] = _container_to_api(
            config=self._config,
            image=container.image,
            entrypoint=container.entrypoint,
            command=container.command,
            working_dir=container.working_dir,
            http=container.http,
            env=container.env,
            volumes=container.volumes,
            secret_env=container.secret_env,
            secret_files=container.secret_files,
            disk_volumes=container.disk_volumes,
            tty=container.tty,
        )
        payload["container"]["resources"] = _resources_to_api(container.resources)
        payload["scheduler_enabled"] = scheduler_enabled
        auth = await self._config._api_auth()
        async with self._core.request("POST", url, json=payload, auth=auth) as resp:
            res = await resp.json()
            return _job_description_from_api(res, self._parse)

    async def start(
        self,
        *,
        image: RemoteImage,
        preset_name: str,
        cluster_name: Optional[str] = None,
        org_name: Optional[str] = None,
        entrypoint: Optional[str] = None,
        command: Optional[str] = None,
        working_dir: Optional[str] = None,
        http: Optional[HTTPPort] = None,
        env: Optional[Mapping[str, str]] = None,
        volumes: Sequence[Volume] = (),
        secret_env: Optional[Mapping[str, URL]] = None,
        secret_files: Sequence[SecretFile] = (),
        disk_volumes: Sequence[DiskVolume] = (),
        tty: bool = False,
        shm: bool = False,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
        description: Optional[str] = None,
        pass_config: bool = False,
        wait_for_jobs_quota: bool = False,
        schedule_timeout: Optional[float] = None,
        restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
        life_span: Optional[float] = None,
        privileged: bool = False,
    ) -> JobDescription:
        url = (self._config.api_url / "jobs").with_query("from_preset")
        container_payload = _container_to_api(
            config=self._config,
            image=image,
            entrypoint=entrypoint,
            command=command,
            working_dir=working_dir,
            http=http,
            env=env,
            volumes=volumes,
            secret_env=secret_env,
            secret_files=secret_files,
            disk_volumes=disk_volumes,
            tty=tty,
            shm=shm,
        )
        payload = _job_to_api(
            cluster_name=cluster_name or self._config.cluster_name,
            name=name,
            preset_name=preset_name,
            tags=tags,
            description=description,
            pass_config=pass_config,
            wait_for_jobs_quota=wait_for_jobs_quota,
            schedule_timeout=schedule_timeout,
            restart_policy=restart_policy,
            life_span=life_span,
            privileged=privileged,
            org_name=org_name,
        )
        payload.update(**container_payload)
        auth = await self._config._api_auth()
        async with self._core.request("POST", url, json=payload, auth=auth) as resp:
            res = await resp.json()
            return _job_description_from_api(res, self._parse)

    @asyncgeneratorcontextmanager
    async def list(
        self,
        *,
        statuses: Iterable[JobStatus] = (),
        name: str = "",
        tags: Iterable[str] = (),
        owners: Iterable[str] = (),
        since: Optional[datetime] = None,
        until: Optional[datetime] = None,
        reverse: bool = False,
        limit: Optional[int] = None,
        cluster_name: Optional[str] = None,
        _materialized: Optional[bool] = None,
        _being_dropped: Optional[bool] = False,
        _logs_removed: Optional[bool] = False,
    ) -> AsyncIterator[JobDescription]:
        url = self._config.api_url / "jobs"
        headers = {"Accept": "application/x-ndjson"}
        params: MultiDict[str] = MultiDict()
        for status in statuses:
            params.add("status", status.value)
        if name:
            params.add("name", name)
        for owner in owners:
            params.add("owner", owner)
        for tag in tags:
            params.add("tag", tag)
        if since:
            if since.tzinfo is None:
                # Interpret naive datetime object as local time.
                since = since.astimezone(timezone.utc)
            params.add("since", since.isoformat())
        if until:
            if until.tzinfo is None:
                until = until.astimezone(timezone.utc)
            params.add("until", until.isoformat())
        if cluster_name is None:
            cluster_name = self._config.cluster_name
        params["cluster_name"] = cluster_name
        if reverse:
            params.add("reverse", "1")
        if limit is not None:
            params.add("limit", str(limit))
        if _materialized is not None:
            params.add("materialized", str(_materialized))
        if _being_dropped is not None:
            params.add("being_dropped", str(_being_dropped))
        if _logs_removed is not None:
            params.add("logs_removed", str(_logs_removed))
        auth = await self._config._api_auth()
        async with self._core.request(
            "GET", url, headers=headers, params=params, auth=auth
        ) as resp:
            if resp.headers.get("Content-Type", "").startswith("application/x-ndjson"):
                async for line in resp.content:
                    server_message = json.loads(line)
                    if "error" in server_message:
                        raise NDJSONError(server_message["error"])
                    yield _job_description_from_api(server_message, self._parse)
            else:
                ret = await resp.json()
                for j in ret["jobs"]:
                    yield _job_description_from_api(j, self._parse)

    async def kill(self, id: str) -> None:
        url = self._config.api_url / "jobs" / id
        auth = await self._config._api_auth()
        async with self._core.request("DELETE", url, auth=auth):
            # an error is raised for status >= 400
            return None  # 201 status code

    async def bump_life_span(self, id: str, additional_life_span: float) -> None:
        url = self._config.api_url / "jobs" / id / "max_run_time_minutes"
        payload = {
            "additional_max_run_time_minutes": int(additional_life_span // 60),
        }
        auth = await self._config._api_auth()
        async with self._core.request("PUT", url, json=payload, auth=auth):
            # an error is raised for status >= 400
            return None  # 201 status code

    @asyncgeneratorcontextmanager
    async def monitor(
        self,
        id: str,
        *,
        cluster_name: Optional[str] = None,
        since: Optional[datetime] = None,
        timestamps: bool = False,
        separator: Optional[str] = None,
        debug: bool = False,
    ) -> AsyncIterator[bytes]:
        url = self._get_monitoring_url(cluster_name) / id / "log"
        if since is not None:
            if since.tzinfo is None:
                # Interpret naive datetime object as local time.
                since = since.astimezone(timezone.utc)
            url = url.update_query(since=since.isoformat())
        if timestamps:
            url = url.update_query(timestamps="true")
        if separator is not None:
            url = url.update_query(separator=separator)
        if debug:
            url = url.update_query(debug="true")
        timeout = attr.evolve(self._core.timeout, sock_read=None)
        auth = await self._config._api_auth()
        async with self._core.request(
            "GET",
            url,
            headers={"Accept-Encoding": "identity"},
            timeout=timeout,
            auth=auth,
        ) as resp:
            async for data in resp.content.iter_any():
                yield data

    async def status(self, id: str) -> JobDescription:
        url = self._config.api_url / "jobs" / id
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, auth=auth) as resp:
            ret = await resp.json()
            return _job_description_from_api(ret, self._parse)

    @asyncgeneratorcontextmanager
    async def top(
        self, id: str, *, cluster_name: Optional[str] = None
    ) -> AsyncIterator[JobTelemetry]:
        url = self._get_monitoring_url(cluster_name) / id / "top"
        auth = await self._config._api_auth()
        try:
            received_any = False
            async with self._core.ws_connect(url, auth=auth) as ws:
                async for resp in ws:
                    yield _job_telemetry_from_api(resp.json())
                    received_any = True
            if not received_any:
                raise ValueError(f"Job is not running. Job Id = {id}")
        except WSServerHandshakeError as e:
            if e.status == 400:
                raise ValueError(f"Job not found. Job Id = {id}")
            raise

    async def save(
        self,
        id: str,
        image: RemoteImage,
        *,
        progress: Optional[AbstractDockerImageProgress] = None,
        cluster_name: Optional[str] = None,
    ) -> None:
        if not image._is_in_neuro_registry:
            raise ValueError(f"Image `{image}` must be in the neuro registry")
        if progress is None:
            progress = _DummyProgress()

        payload = {"container": {"image": image.as_docker_url()}}
        url = self._get_monitoring_url(cluster_name) / id / "save"

        auth = await self._config._api_auth()
        timeout = attr.evolve(self._core.timeout, sock_read=None)
        # `self._code.request` implicitly sets `total=3 * 60`
        # unless `sock_read is None`
        async with self._core.request(
            "POST", url, json=payload, timeout=timeout, auth=auth
        ) as resp:
            # first, we expect exactly two docker-commit messages
            progress.save(ImageProgressSave(id, image))

            chunk_1 = await resp.content.readline()
            data_1 = _parse_commit_started_chunk(id, _load_chunk(chunk_1), self._parse)
            progress.commit_started(data_1)

            chunk_2 = await resp.content.readline()
            data_2 = _parse_commit_finished_chunk(id, _load_chunk(chunk_2))
            progress.commit_finished(data_2)

            # then, we expect stream for docker-push
            src = LocalImage(f"{image.owner}/{image.name}", image.tag)
            progress.push(ImageProgressPush(src, dst=image))
            async for chunk in resp.content:
                obj = _load_chunk(chunk)
                push_step = _try_parse_image_progress_step(obj, image.tag)
                if push_step:
                    progress.step(push_step)

    @asynccontextmanager
    async def port_forward(
        self,
        id: str,
        local_port: int,
        job_port: int,
        *,
        no_key_check: bool = False,
        cluster_name: Optional[str] = None,
    ) -> AsyncIterator[None]:
        srv = await asyncio.start_server(
            partial(
                self._port_forward, id=id, job_port=job_port, cluster_name=cluster_name
            ),
            "localhost",
            local_port,
        )
        try:
            yield
        finally:
            srv.close()
            await srv.wait_closed()

    async def _port_forward(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        id: str,
        job_port: int,
        *,
        cluster_name: Optional[str] = None,
    ) -> None:
        try:
            loop = asyncio.get_event_loop()
            url = self._get_monitoring_url(cluster_name)
            url = url / id / "port_forward" / str(job_port)
            auth = await self._config._api_auth()
            ws = await self._core._session.ws_connect(
                url,
                headers={"Authorization": auth},
                timeout=None,  # type: ignore
                receive_timeout=None,
                heartbeat=30,
            )
            tasks = []
            tasks.append(loop.create_task(self._port_reader(ws, writer)))
            tasks.append(loop.create_task(self._port_writer(ws, reader)))
            try:
                await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            finally:
                for task in tasks:
                    if not task.done():
                        task.cancel()
                        with suppress(asyncio.CancelledError):
                            await task
                writer.close()
                await writer.wait_closed()
                await ws.close()
        except asyncio.CancelledError:
            raise
        except WSServerHandshakeError as e:
            if e.headers and "X-Error" in e.headers:
                log.error(f"Error during port-forwarding: {e.headers['X-Error']}")
            log.exception("Unhandled exception during port-forwarding")
            writer.close()
        except Exception:
            log.exception("Unhandled exception during port-forwarding")
            writer.close()

    async def _port_reader(
        self, ws: aiohttp.ClientWebSocketResponse, writer: asyncio.StreamWriter
    ) -> None:
        async for msg in ws:
            assert msg.type == aiohttp.WSMsgType.BINARY
            writer.write(msg.data)
            await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def _port_writer(
        self, ws: aiohttp.ClientWebSocketResponse, reader: asyncio.StreamReader
    ) -> None:
        while True:
            data = await reader.read(4 * 1024 * 1024)
            if not data:
                # EOF
                break
            await ws.send_bytes(data)

    @asynccontextmanager
    async def attach(
        self,
        id: str,
        *,
        tty: bool = False,
        stdin: bool = False,
        stdout: bool = False,
        stderr: bool = False,
        cluster_name: Optional[str] = None,
    ) -> AsyncIterator[StdStream]:
        url = self._get_monitoring_url(cluster_name) / id / "attach"
        url = url.with_query(
            tty=str(int(tty)),
            stdin=str(int(stdin)),
            stdout=str(int(stdout)),
            stderr=str(int(stderr)),
        )
        auth = await self._config._api_auth()

        try:
            ws = await self._core._session.ws_connect(
                url,
                headers={
                    "Authorization": auth,
                    aiohttp.hdrs.SEC_WEBSOCKET_PROTOCOL: "v2.channels.neu.ro",
                },
                timeout=None,  # type: ignore
                receive_timeout=None,
                heartbeat=30,
            )
        except aiohttp.ClientResponseError as ex:
            if ex.status == 404:
                raise ResourceNotFound(f"Job {id!r} is not running")
            raise

        try:
            yield StdStream(ws)
        finally:
            await ws.close()

    @asynccontextmanager
    async def exec(
        self,
        id: str,
        cmd: str,
        *,
        tty: bool = False,
        stdin: bool = False,
        stdout: bool = False,
        stderr: bool = False,
        cluster_name: Optional[str] = None,
    ) -> AsyncIterator[StdStream]:
        url = self._get_monitoring_url(cluster_name) / id / "exec"
        url = url.with_query(
            cmd=cmd,
            tty=str(int(tty)),
            stdin=str(int(stdin)),
            stdout=str(int(stdout)),
            stderr=str(int(stderr)),
        )
        auth = await self._config._api_auth()

        try:
            ws = await self._core._session.ws_connect(
                url,
                headers={"Authorization": auth},
                timeout=None,  # type: ignore
                receive_timeout=None,
                heartbeat=30,
            )
        except aiohttp.ClientResponseError as ex:
            if ex.status == 404:
                raise ResourceNotFound(f"Job {id!r} is not running")
            raise

        try:
            yield StdStream(ws)
        finally:
            await ws.close()

    async def send_signal(self, id: str, *, cluster_name: Optional[str] = None) -> None:
        url = self._get_monitoring_url(cluster_name) / id / "kill"
        auth = await self._config._api_auth()
        async with self._core.request("POST", url, auth=auth) as resp:
            resp

    async def get_capacity(
        self, *, cluster_name: Optional[str] = None
    ) -> Mapping[str, int]:
        url = self._get_monitoring_url(cluster_name) / "capacity"
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, auth=auth) as resp:
            return await resp.json()


#  ############## Internal helpers ###################


def _load_chunk(chunk: bytes) -> Dict[str, Any]:
    return json.loads(chunk.decode())


def _parse_commit_started_chunk(
    job_id: str, obj: Dict[str, Any], parse: Parser
) -> ImageCommitStarted:
    _raise_for_invalid_commit_chunk(obj, expect_started=True)
    details_json = obj.get("details", {})
    image = details_json.get("image")
    if not image:
        error_details = {"message": "Missing required details: 'image'"}
        raise DockerError(400, error_details)
    return ImageCommitStarted(job_id, parse.remote_image(image))


def _parse_commit_finished_chunk(
    job_id: str, obj: Dict[str, Any]
) -> ImageCommitFinished:
    _raise_for_invalid_commit_chunk(obj, expect_started=False)
    return ImageCommitFinished(job_id)


def _raise_for_invalid_commit_chunk(obj: Dict[str, Any], expect_started: bool) -> None:
    _raise_on_error_chunk(obj)
    if "status" not in obj.keys():
        error_details = {"message": 'Missing required field: "status"'}
        raise DockerError(400, error_details)
    status = obj["status"]
    expected = "CommitStarted" if expect_started else "CommitFinished"
    if status != expected:
        error_details = {
            "message": f"Invalid commit status: '{status}', expecting: '{expected}'"
        }
        raise DockerError(400, error_details)


def _resources_to_api(resources: Resources) -> Dict[str, Any]:
    value: Dict[str, Any] = {
        "memory_mb": resources.memory_mb,
        "cpu": resources.cpu,
        "shm": resources.shm,
    }
    if resources.gpu:
        value["gpu"] = resources.gpu
        value["gpu_model"] = resources.gpu_model
    if resources.tpu_type:
        assert resources.tpu_software_version
        value["tpu"] = {
            "type": resources.tpu_type,
            "software_version": resources.tpu_software_version,
        }
    return value


def _resources_from_api(data: Dict[str, Any]) -> Resources:
    tpu_type = tpu_software_version = None
    if "tpu" in data:
        tpu = data["tpu"]
        tpu_type = tpu["type"]
        tpu_software_version = tpu["software_version"]
    return Resources(
        memory_mb=data["memory_mb"],
        cpu=data["cpu"],
        shm=data.get("shm", True),
        gpu=data.get("gpu", None),
        gpu_model=data.get("gpu_model", None),
        tpu_type=tpu_type,
        tpu_software_version=tpu_software_version,
    )


def _http_port_to_api(port: HTTPPort) -> Dict[str, Any]:
    return {"port": port.port, "requires_auth": port.requires_auth}


def _http_port_from_api(data: Dict[str, Any]) -> HTTPPort:
    return HTTPPort(
        port=data.get("port", -1), requires_auth=data.get("requires_auth", False)
    )


def _container_from_api(data: Dict[str, Any], parse: Parser) -> Container:
    try:
        image = parse.remote_image(data["image"])
    except ValueError:
        image = RemoteImage.new_external_image(name=INVALID_IMAGE_NAME)

    return Container(
        image=image,
        resources=_resources_from_api(data["resources"]),
        entrypoint=data.get("entrypoint", None),
        command=data.get("command", None),
        working_dir=data.get("working_dir"),
        http=_http_port_from_api(data["http"]) if "http" in data else None,
        env=data.get("env", dict()),
        volumes=[_volume_from_api(v) for v in data.get("volumes", [])],
        secret_env={name: URL(val) for name, val in data.get("secret_env", {}).items()},
        secret_files=[_secret_file_from_api(v) for v in data.get("secret_volumes", [])],
        disk_volumes=[_disk_volume_from_api(v) for v in data.get("disk_volumes", [])],
        tty=data.get("tty", False),
    )


def _container_to_api(
    config: Config,
    image: RemoteImage,
    entrypoint: Optional[str] = None,
    command: Optional[str] = None,
    working_dir: Optional[str] = None,
    http: Optional[HTTPPort] = None,
    env: Optional[Mapping[str, str]] = None,
    volumes: Sequence[Volume] = (),
    secret_env: Optional[Mapping[str, URL]] = None,
    secret_files: Sequence[SecretFile] = (),
    disk_volumes: Sequence[DiskVolume] = (),
    tty: bool = False,
    shm: bool = False,
) -> Dict[str, Any]:
    primitive: Dict[str, Any] = {"image": image.as_docker_url()}
    if shm:
        primitive["resources"] = {"shm": shm}
    if entrypoint:
        primitive["entrypoint"] = entrypoint
    if command:
        primitive["command"] = command
    if working_dir:
        primitive["working_dir"] = working_dir
    if http:
        primitive["http"] = _http_port_to_api(http)
    if env:
        primitive["env"] = env
    if volumes:
        primitive["volumes"] = [_volume_to_api(v, config) for v in volumes]
    if secret_env:
        primitive["secret_env"] = {
            k: str(
                normalize_secret_uri(
                    v, config.username, config.cluster_name, config.org_name
                )
            )
            for k, v in secret_env.items()
        }
    if secret_files:
        primitive["secret_volumes"] = [
            _secret_file_to_api(v, config) for v in secret_files
        ]
    if disk_volumes:
        primitive["disk_volumes"] = [
            _disk_volume_to_api(v, config) for v in disk_volumes
        ]
    if tty:
        primitive["tty"] = True
    return primitive


def _calc_status(stat: str) -> JobStatus:
    # Forward-compatible support for CANCELLED status
    try:
        return JobStatus(stat)
    except ValueError:
        return JobStatus.UNKNOWN


def _job_status_item_from_api(res: Dict[str, Any]) -> JobStatusItem:
    return JobStatusItem(
        status=_calc_status(res.get("status", "unknown")),
        transition_time=_parse_datetime(res["transition_time"]),
        reason=res.get("reason", ""),
        description=res.get("description", ""),
        exit_code=res.get("exit_code"),
    )


def _job_description_from_api(res: Dict[str, Any], parse: Parser) -> JobDescription:
    # TODO y.s.: maybe, catch KeyErrors and re-raise with an error message like
    #   "SDK and API has incompatible versions: {key} was not found in the API responce"
    container = _container_from_api(res["container"], parse)
    owner = res["owner"]
    cluster_name = res["cluster_name"]
    name = res.get("name")
    tags = res.get("tags", ())
    description = res.get("description")
    history = JobStatusHistory(
        # Forward-compatible support for CANCELLED status
        status=_calc_status(res["history"].get("status", "unknown")),
        reason=res["history"].get("reason", ""),
        restarts=res["history"].get("restarts", 0),
        description=res["history"].get("description", ""),
        created_at=_parse_datetime(res["history"].get("created_at")),
        started_at=_parse_datetime(res["history"].get("started_at")),
        finished_at=_parse_datetime(res["history"].get("finished_at")),
        run_time_seconds=res["history"].get("run_time_seconds"),
        exit_code=res["history"].get("exit_code"),
        transitions=[
            _job_status_item_from_api(item_raw) for item_raw in res.get("statuses", [])
        ],
    )
    http_url = URL(res.get("http_url", ""))
    http_url_named = URL(res.get("http_url_named", ""))
    internal_hostname = res.get("internal_hostname", None)
    internal_hostname_named = res.get("internal_hostname_named", None)
    restart_policy = JobRestartPolicy(res.get("restart_policy", JobRestartPolicy.NEVER))
    max_run_time_minutes = res.get("max_run_time_minutes")
    life_span = (
        max_run_time_minutes * 60.0 if max_run_time_minutes is not None else None
    )
    # TODO: this change requires a new release of API with
    # https://github.com/neuro-inc/platform-api/pull/1770 merged
    total_price_credits = Decimal(res["total_price_credits"])
    price_credits_per_hour = Decimal(res["price_credits_per_hour"])
    return JobDescription(
        status=_calc_status(res["status"]),
        id=res["id"],
        owner=owner,
        cluster_name=cluster_name,
        org_name=res.get("org_name"),
        history=history,
        container=container,
        scheduler_enabled=res["scheduler_enabled"],
        preemptible_node=res.get("preemptible_node", False),
        pass_config=res["pass_config"],
        name=name,
        tags=tags,
        description=description,
        http_url=http_url_named or http_url,
        internal_hostname=internal_hostname,
        internal_hostname_named=internal_hostname_named,
        uri=URL(res["uri"]),
        restart_policy=restart_policy,
        life_span=life_span,
        schedule_timeout=res.get("schedule_timeout", None),
        preset_name=res.get("preset_name"),
        total_price_credits=total_price_credits,
        price_credits_per_hour=price_credits_per_hour,
        _internal=JobDescriptionInternal(
            materialized=res.get("materialized", False),
            being_dropped=res.get("being_dropped", False),
            logs_removed=res.get("logs_removed", False),
        ),
    )


def _job_to_api(
    cluster_name: str,
    name: Optional[str] = None,
    preset_name: Optional[str] = None,
    tags: Sequence[str] = (),
    description: Optional[str] = None,
    pass_config: bool = False,
    wait_for_jobs_quota: bool = False,
    schedule_timeout: Optional[float] = None,
    restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
    life_span: Optional[float] = None,
    privileged: bool = False,
    org_name: Optional[str] = None,
) -> Dict[str, Any]:
    primitive: Dict[str, Any] = {"pass_config": pass_config}
    if name:
        primitive["name"] = name
    if preset_name:
        primitive["preset_name"] = preset_name
    if tags:
        primitive["tags"] = tags
    if description:
        primitive["description"] = description
    if schedule_timeout:
        primitive["schedule_timeout"] = schedule_timeout
    if restart_policy != JobRestartPolicy.NEVER:
        primitive["restart_policy"] = str(restart_policy)
    if life_span is not None:
        primitive["max_run_time_minutes"] = int(life_span // 60)
    if wait_for_jobs_quota:
        primitive["wait_for_jobs_quota"] = wait_for_jobs_quota
    if privileged:
        primitive["privileged"] = privileged
    if org_name:
        primitive["org_name"] = org_name
    primitive["cluster_name"] = cluster_name
    return primitive


def _job_telemetry_from_api(value: Dict[str, Any]) -> JobTelemetry:
    return JobTelemetry(
        cpu=value["cpu"],
        memory=value["memory"],
        timestamp=value["timestamp"],
        gpu_duty_cycle=value.get("gpu_duty_cycle"),
        gpu_memory=value.get("gpu_memory"),
    )


def _volume_to_api(volume: Volume, config: Config) -> Dict[str, Any]:
    uri = normalize_storage_path_uri(
        volume.storage_uri, config.username, config.cluster_name, config.org_name
    )
    resp: Dict[str, Any] = {
        "src_storage_uri": str(uri),
        "dst_path": volume.container_path,
        "read_only": bool(volume.read_only),
    }
    return resp


def _secret_file_to_api(secret_file: SecretFile, config: Config) -> Dict[str, Any]:
    uri = normalize_secret_uri(
        secret_file.secret_uri, config.username, config.cluster_name, config.org_name
    )
    return {
        "src_secret_uri": str(uri),
        "dst_path": secret_file.container_path,
    }


def _disk_volume_to_api(volume: DiskVolume, config: Config) -> Dict[str, Any]:
    uri = normalize_disk_uri(
        volume.disk_uri, config.username, config.cluster_name, config.org_name
    )
    resp: Dict[str, Any] = {
        "src_disk_uri": str(uri),
        "dst_path": volume.container_path,
        "read_only": bool(volume.read_only),
    }
    return resp


def _volume_from_api(data: Dict[str, Any]) -> Volume:
    storage_uri = URL(data["src_storage_uri"])
    container_path = data["dst_path"]
    read_only = data.get("read_only", True)
    return Volume(
        storage_uri=storage_uri, container_path=container_path, read_only=read_only
    )


def _secret_file_from_api(data: Dict[str, Any]) -> SecretFile:
    secret_uri = URL(data["src_secret_uri"])
    container_path = data["dst_path"]
    return SecretFile(secret_uri, container_path)


def _disk_volume_from_api(data: Dict[str, Any]) -> DiskVolume:
    disk_uri = URL(data["src_disk_uri"])
    container_path = data["dst_path"]
    read_only = data.get("read_only", True)
    return DiskVolume(disk_uri, container_path, read_only)


@overload
def _parse_datetime(dt: str) -> datetime:
    ...


@overload
def _parse_datetime(dt: Optional[str]) -> Optional[datetime]:
    ...


def _parse_datetime(dt: Optional[str]) -> Optional[datetime]:
    if dt is None:
        return None
    return isoparse(dt)



from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, Optional, Sequence

from aiohttp.web import HTTPCreated, HTTPNoContent
from yarl import URL

from ._admin import _Admin
from ._config import Config
from ._core import _Core
from ._errors import ClientError, NotSupportedError
from ._rewrite import rewrite_module
from ._utils import NoPublicConstructor


@rewrite_module
class Action(str, Enum):
    READ = "read"
    WRITE = "write"
    MANAGE = "manage"


@rewrite_module
@dataclass(frozen=True)
class Permission:
    uri: URL
    action: Action


@rewrite_module
@dataclass(frozen=True)
class Share:
    user: str
    permission: Permission


@rewrite_module
@dataclass(frozen=True)
class Quota:
    credits: Optional[Decimal] = None
    total_running_jobs: Optional[int] = None


@rewrite_module
class Users(metaclass=NoPublicConstructor):
    def __init__(self, core: _Core, config: Config, admin: _Admin) -> None:
        self._core = core
        self._config = config
        self._admin = admin

    async def get_quota(self) -> Quota:
        try:
            ret = await self._admin.get_cluster_user(
                cluster_name=self._config.cluster_name,
                org_name=self._config.org_name,
                user_name=self._config.username,
            )
        except NotSupportedError:
            # FOSS configuration without admin service and limits
            return Quota(credits=None, total_running_jobs=None)
        return Quota(
            credits=ret.balance.credits,
            total_running_jobs=ret.quota.total_running_jobs,
        )

    async def get_acl(
        self, user: str, scheme: Optional[str] = None, *, uri: Optional[URL] = None
    ) -> Sequence[Permission]:
        url = self._get_user_url(user) / "permissions"
        if scheme:
            if uri is not None:
                raise ValueError("Conflicting arguments 'uri' and 'scheme'")
            uri = URL.build(scheme=scheme)
        params = {"uri": str(uri)} if uri is not None else {}
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, params=params, auth=auth) as resp:
            payload = await resp.json()
        ret = []
        for item in payload:
            uri = URL(item["uri"])
            action = Action(item["action"])
            ret.append(Permission(uri, action))
        return ret

    async def get_shares(
        self, user: str, scheme: Optional[str] = None, *, uri: Optional[URL] = None
    ) -> Sequence[Share]:
        url = self._get_user_url(user) / "permissions" / "shared"
        if scheme:
            if uri is not None:
                raise ValueError("Conflicting arguments 'uri' and 'scheme'")
            uri = URL.build(scheme=scheme)
        params = {"uri": str(uri)} if uri is not None else {}
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, params=params, auth=auth) as resp:
            payload = await resp.json()
        ret = []
        for item in payload:
            uri = URL(item["uri"])
            action = Action(item["action"])
            ret.append(Share(item["user"], Permission(uri, action)))
        return ret

    async def get_subroles(
        self,
        user: str,
    ) -> Sequence[str]:
        url = self._get_user_url(user) / "subroles"
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, auth=auth) as resp:
            payload = await resp.json()
        return payload["subroles"]

    async def share(self, user: str, permission: Permission) -> Permission:
        url = self._get_user_url(user) / "permissions"
        payload = [_permission_to_api(permission)]
        auth = await self._config._api_auth()
        async with self._core.request("POST", url, json=payload, auth=auth) as resp:
            if resp.status != HTTPCreated.status_code:
                raise ClientError("Server return unexpected result.")
            payload = await resp.json()
            perm = payload[0]
            return Permission(
                uri=URL(perm["uri"]),
                action=Action(perm["action"]),
            )

    async def revoke(self, user: str, uri: URL) -> None:
        url = self._get_user_url(user) / "permissions"
        auth = await self._config._api_auth()
        async with self._core.request(
            "DELETE", url, params={"uri": str(uri)}, auth=auth
        ) as resp:
            #  TODO: server part contain TODO record for returning more then
            #  HTTPNoContent, this part must me refactored then
            if resp.status != HTTPNoContent.status_code:
                raise ClientError(f"Server return unexpected result: {resp.status}.")
        return None

    async def add(self, role_name: str) -> None:
        url = self._config.api_url / "users"
        auth = await self._config._api_auth()
        async with self._core.request(
            "POST", url, json={"name": role_name}, auth=auth
        ) as resp:
            if resp.status != HTTPCreated.status_code:
                raise ClientError(f"Server return unexpected result: {resp.status}.")
        return None

    async def remove(self, role_name: str) -> None:
        url = self._get_user_url(role_name)
        auth = await self._config._api_auth()
        async with self._core.request("DELETE", url, auth=auth) as resp:
            if resp.status != HTTPNoContent.status_code:
                raise ClientError(f"Server return unexpected result: {resp.status}.")
        return None

    def _get_user_url(self, user: str) -> URL:
        if ":" in user:
            raise ValueError(f"Invalid name: {user!r}")
        return self._config.api_url / "users" / user.replace("/", ":")


def _permission_to_api(perm: Permission) -> Dict[str, Any]:
    primitive: Dict[str, Any] = {"uri": str(perm.uri), "action": perm.action.value}
    return primitive



import os
import sys
from typing import Any, AsyncIterator, Callable, Dict, Iterator
from unittest import mock

import pytest
from aiodocker.exceptions import DockerError
from aiohttp import web
from aiohttp.hdrs import LINK
from yarl import URL

from neuro_sdk import (
    AuthorizationError,
    Client,
    LocalImage,
    RemoteImage,
    Tag,
    TagOption,
)
from neuro_sdk._parsing_utils import _get_url_authority, _ImageNameParser

from tests import _TestServerFactory

if sys.version_info < (3, 8):
    pytest.skip("Async mocks require Python 3.8+", allow_module_level=True)

_MakeClient = Callable[..., Client]


@pytest.fixture()
def patch_docker_host() -> Iterator[None]:
    with mock.patch.dict(os.environ, values={"DOCKER_HOST": "http://localhost:45678"}):
        yield


class TestImageParser:
    parser = _ImageNameParser(
        default_user="alice",
        default_cluster="default",
        default_org=None,
        registry_urls={
            "default": URL("https://reg.neu.ro"),
            "another": URL("https://example.org"),
        },
    )

    @pytest.mark.parametrize(
        "image",
        [
            "image://default/me/ubuntu:v10.04",
            "image://another/me/ubuntu:v10.04",
            "image:ubuntu:v10.04",
            "image:///ubuntu:v10.04",
            "image:ubuntu:v10.04",
            "ubuntu:v10.04",
        ],
    )
    def test_has_tag_ok(self, image: str) -> None:
        assert self.parser.has_tag(image)

    def test_has_tag_no_tag(self) -> None:
        image = "ubuntu"
        assert not self.parser.has_tag(image)

    def test_has_tag_no_tag_with_slash(self) -> None:
        image = "library/ubuntu"
        assert not self.parser.has_tag(image)

    def test_has_tag_empty_tag(self) -> None:
        image = "ubuntu:"
        with pytest.raises(ValueError, match="empty tag"):
            self.parser.has_tag(image)

    def test_has_tag_empty_tag_with_slash(self) -> None:
        image = "library/ubuntu:"
        with pytest.raises(ValueError, match="empty tag"):
            self.parser.has_tag(image)

    def test_has_tag_empty_image_name(self) -> None:
        image = ":latest"
        with pytest.raises(ValueError, match="empty name"):
            self.parser.has_tag(image)

    def test_has_tag_too_many_tags(self) -> None:
        image = "ubuntu:v10.04:latest"
        with pytest.raises(ValueError, match="too many tags"):
            self.parser.has_tag(image)

    def test_has_tag_lstrip(self) -> None:
        image = "image:game:latest"
        assert self.parser.has_tag(image)
        image = "image:game:mega"
        assert self.parser.has_tag(image)
        image = "image:game:v2.0:latest"
        with pytest.raises(ValueError, match="too many tags"):
            assert self.parser.has_tag(image)

    @pytest.mark.parametrize(
        "registry_url",
        ["http://reg.neu.ro", "https://reg.neu.ro", "https://reg.neu.ro/bla/bla"],
    )
    def test_get_registry_hostname(self, registry_url: str) -> None:
        parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL(registry_url)},
        )
        assert parser._registries == {"test-cluster": "reg.neu.ro"}

    @pytest.mark.parametrize(
        "registry_url", ["http://reg.neu.ro:5000", "http://reg.neu.ro:5000/bla/bla"]
    )
    def test_get_registry_hostname_with_port(self, registry_url: str) -> None:
        parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL(registry_url)},
        )
        assert parser._registries == {"test-cluster": "reg.neu.ro:5000"}

    @pytest.mark.parametrize(
        "registry_url",
        ["", "reg.neu.ro", "reg.neu.ro:5000", "https://", "https:///bla/bla"],
    )
    def test_get_registry_hostname__bad_url_empty_hostname(
        self, registry_url: str
    ) -> None:
        with pytest.raises(ValueError, match="Empty hostname in registry URL"):
            _ImageNameParser(
                default_user="alice",
                default_cluster="test-cluster",
                default_org=None,
                registry_urls={"test-cluster": URL(registry_url)},
            )

    def test_split_image_name_no_tag(self) -> None:
        splitted = self.parser._split_image_name("ubuntu", "latest")
        assert splitted == ("ubuntu", "latest")

    def test_split_image_name_with_tag(self) -> None:
        splitted = self.parser._split_image_name("ubuntu:v10.04", "latest")
        assert splitted == ("ubuntu", "v10.04")

    def test_split_image_name_empty_tag(self) -> None:
        with pytest.raises(ValueError, match="empty tag"):
            self.parser._split_image_name("ubuntu:", "latest")

    def test_split_image_name_two_tags(self) -> None:
        with pytest.raises(ValueError, match="too many tags"):
            self.parser._split_image_name("ubuntu:v10.04:LTS", "latest")

    def test_split_image_name_with_registry_port_no_tag(self) -> None:
        splitted = self.parser._split_image_name("localhost:5000/ubuntu", "latest")
        assert splitted == ("localhost:5000/ubuntu", "latest")

    def test_split_image_name_with_registry_port_with_tag(self) -> None:
        splitted = self.parser._split_image_name(
            "localhost:5000/ubuntu:v10.04", "latest"
        )
        assert splitted == ("localhost:5000/ubuntu", "v10.04")

    def test_split_image_name_with_registry_port_two_tags(self) -> None:
        with pytest.raises(ValueError, match="too many tags"):
            self.parser._split_image_name("localhost:5000/ubuntu:v10.04:LTS", "latest")

    def test_split_image_name_with_registry_port_empty_tag(self) -> None:
        with pytest.raises(ValueError, match="empty tag"):
            self.parser._split_image_name("localhost:5000/ubuntu:", "latest")

    def test_split_image_name_with_registry_port_slash_in_tag(self) -> None:
        with pytest.raises(ValueError, match="invalid tag"):
            self.parser._split_image_name("localhost:5000/ubuntu:v10/04", "latest")

    # public method: parse_local

    @pytest.mark.parametrize(
        "url",
        [
            "image://",
            "image:///",
            "image://default/bob",
            "image://default/bob/",
        ],
    )
    def test_parse_as_neuro_image__no_image_name(self, url: str) -> None:
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_local_image_empty_fail(self) -> None:
        image = ""
        with pytest.raises(ValueError, match="empty image name"):
            self.parser.parse_as_local_image(image)

    def test_parse_as_local_image_dash_fail(self) -> None:
        image = "-zxc"
        with pytest.raises(ValueError, match="image cannot start with dash"):
            self.parser.parse_as_local_image(image)

    def test_parse_as_local_image_with_image_scheme_fail(self) -> None:
        image = "image://default/ubuntu"
        with pytest.raises(
            ValueError, match="scheme 'image://' is not allowed for local images"
        ):
            self.parser.parse_as_local_image(image)

    def test_parse_as_local_image_with_other_scheme_ok(self) -> None:
        image = "http://ubuntu"
        parsed = self.parser.parse_as_local_image(image)
        # instead of parser, the docker client will fail
        assert parsed == LocalImage(name="http://ubuntu", tag="latest")
        assert self.parser.parse_as_local_image(str(parsed)) == parsed

    def test_parse_as_local_image_no_tag(self) -> None:
        image = "ubuntu"
        parsed = self.parser.parse_as_local_image(image)
        assert parsed == LocalImage(name="ubuntu", tag="latest")

    def test_parse_as_local_image_with_tag(self) -> None:
        image = "ubuntu:v10.04"
        parsed = self.parser.parse_as_local_image(image)
        assert parsed == LocalImage(name="ubuntu", tag="v10.04")
        assert self.parser.parse_as_local_image(str(parsed)) == parsed

    def test_parse_as_local_image_special_chars(self) -> None:
        image = "image#%2d?ß:tag#%2d?ß"
        parsed = self.parser.parse_as_local_image(image)
        assert parsed == LocalImage(name="image#%2d?ß", tag="tag#%2d?ß")
        assert self.parser.parse_as_local_image(str(parsed)) == parsed

    def test_parse_as_local_image_2_tag_fail(self) -> None:
        image = "ubuntu:v10.04:LTS"
        with pytest.raises(ValueError, match="too many tags"):
            self.parser.parse_as_local_image(image)

    # public method: parse_remote

    @pytest.mark.parametrize(
        "url",
        [
            "image://default/bob/ubuntu:v10.04?key=value",
            "image://default/bob/ubuntu?key=value",
            "image:///ubuntu?key=value",
            "image:ubuntu?key=value",
            "image:5000?key=value",
            "image://default/bob/ubuntu:v10.04?",
        ],
    )
    def test_parse_as_neuro_image__with_query__fail(self, url: str) -> None:
        with pytest.raises(ValueError, match="Query part is not allowed in image URI"):
            self.parser.parse_as_neuro_image(url)

    @pytest.mark.parametrize(
        "url",
        [
            "image://default/bob/ubuntu:v10.04#fragment",
            "image://default/bob/ubuntu#fragment",
            "image:///ubuntu#fragment",
            "image:ubuntu#fragment",
            "image:5000#fragment",
            "image://default/bob/ubuntu:v10.04#",
        ],
    )
    def test_parse_as_neuro_image__with_fragment__fail(self, url: str) -> None:
        with pytest.raises(
            ValueError, match="Fragment part is not allowed in image URI"
        ):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_neuro_image__with_user__fail(self) -> None:
        url = "image://user@default/bob/ubuntu"
        with pytest.raises(ValueError, match="User is not allowed in image URI"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_neuro_image__with_password__fail(self) -> None:
        url = "image://:password@default/bob/ubuntu"
        with pytest.raises(ValueError, match="Password is not allowed in image URI"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_neuro_image__with_empty_password__fail(self) -> None:
        url = "image://:@default/bob/ubuntu"
        with pytest.raises(ValueError, match="Password is not allowed in image URI"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_neuro_image__with_port__fail(self) -> None:
        url = "image://default:443/bob/ubuntu"
        with pytest.raises(ValueError, match="Port is not allowed in image URI"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_neuro_image_empty__fail(self) -> None:
        image = ""
        with pytest.raises(ValueError, match="empty image name"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_dash__fail(self) -> None:
        image = "-zxc"
        with pytest.raises(ValueError, match="image cannot start with dash"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_fail(self) -> None:
        image = "ubuntu"
        with pytest.raises(
            ValueError, match="scheme 'image:' is required for remote images"
        ):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_invalid_scheme_1_fail(self) -> None:
        image = "ubuntu:latest"
        with pytest.raises(
            ValueError, match="scheme 'image:' is required for remote images"
        ):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_invalid_scheme_2_fail(self) -> None:
        image = "http://ubuntu"
        with pytest.raises(
            ValueError, match="scheme 'image:' is required for remote images"
        ):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_org(
        self,
    ) -> None:
        image = "image://another/test-org/bob/ubuntu:v10.04"
        parser = _ImageNameParser(
            default_user="alice",
            default_cluster="default",
            default_org="test-org",
            registry_urls={
                "default": URL("https://reg.neu.ro"),
                "another": URL("https://example.org"),
            },
        )
        parsed = parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name="test-org",
        )

    def test_parse_as_neuro_image_with_scheme_with_cluster_with_user_with_tag(
        self,
    ) -> None:
        image = "image://another/bob/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_with_cluster_with_user_with_tag_2(
        self,
    ) -> None:
        image = "image://another/bob/library/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_with_cluster_with_user_no_tag(
        self,
    ) -> None:
        image = "image://another/bob/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_with_cluster_with_user_no_tag_2(
        self,
    ) -> None:
        image = "image://another/bob/library/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_unknown_cluster(self) -> None:
        image = "image://unknown-cluster/bob/ubuntu"
        with pytest.raises(
            RuntimeError,
            match="Cluster unknown-cluster doesn't exist "
            "in a list of available clusters",
        ):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_no_slash_no_cluster_no_user_no_tag(
        self,
    ) -> None:
        image = "image:ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_no_slash_no_cluster_no_user_no_tag_2(
        self,
    ) -> None:
        image = "image:library/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="latest",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_no_slash_no_cluster_no_user_with_tag(
        self,
    ) -> None:
        image = "image:ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_no_slash_no_cluster_no_user_with_tag_2(
        self,
    ) -> None:
        image = "image:library/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_no_name_no_tag(
        self,
    ) -> None:
        image = "image:/ubuntu"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_no_tag(self) -> None:
        image = "image:/bob/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_no_tag_2(self) -> None:
        image = "image:/bob/library/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_no_name_with_tag(
        self,
    ) -> None:
        image = "image:/ubuntu:v10.04"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_with_tag(self) -> None:
        image = "image:/bob/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_1_slash_no_cluster_with_tag_2(
        self,
    ) -> None:
        image = "image:/bob/library/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_2_slash_cluster_user_no_name_no_tag_fail(
        self,
    ) -> None:
        image = "image://another/ubuntu"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_2_slash_cluster_no_user_with_tag_fail(
        self,
    ) -> None:
        image = "image://ubuntu:v10.04"
        with pytest.raises(ValueError, match="port can't be converted to integer"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_no_name_no_tag(
        self,
    ) -> None:
        image = "image:///ubuntu"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_no_tag(self) -> None:
        image = "image:///bob/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_no_tag_2(self) -> None:
        image = "image:///bob/library/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_no_name_with_tag(
        self,
    ) -> None:
        image = "image:///ubuntu:v10.04"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_with_tag(self) -> None:
        image = "image:///bob/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_3_slash_no_cluster_with_tag_2(
        self,
    ) -> None:
        image = "image:///bob/library/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_no_name_with_tag(
        self,
    ) -> None:
        image = "image:////ubuntu:v10.04"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_with_tag(self) -> None:
        image = "image:////bob/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_with_tag_2(
        self,
    ) -> None:
        image = "image:////bob/library/ubuntu:v10.04"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_no_name_no_tag(
        self,
    ) -> None:
        image = "image:////ubuntu"
        with pytest.raises(ValueError, match="no image name specified"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_no_tag(self) -> None:
        image = "image:////bob/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_4_slash_no_cluster_no_tag_2(self) -> None:
        image = "image:////bob/library/ubuntu"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="latest",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_with_scheme_special_chars(self) -> None:
        image = "image://another/bob/ubuntu%23%252d%3F%C3%9F:v10.04%23%252d%3F%C3%9F"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu#%2d?ß",
            tag="v10.04#%2d?ß",
            owner="bob",
            cluster_name="another",
            registry="example.org",
            org_name=None,
        )
        assert self.parser.parse_as_neuro_image(str(parsed)) == parsed

    def test_parse_as_neuro_image_no_scheme_no_slash_no_tag_fail(self) -> None:
        image = "ubuntu"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_no_slash_with_tag_fail(self) -> None:
        image = "ubuntu:v10.04"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_1_slash_no_tag_fail(self) -> None:
        image = "library/ubuntu"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_1_slash_with_tag_fail(self) -> None:
        image = "library/ubuntu:v10.04"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_2_slash_no_tag_fail(self) -> None:
        image = "docker.io/library/ubuntu"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_2_slash_with_tag_fail(self) -> None:
        image = "docker.io/library/ubuntu:v10.04"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_no_scheme_3_slash_no_tag_fail(self) -> None:
        image = "something/docker.io/library/ubuntu"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_with_registry_prefix(self) -> None:
        image = self.parser.parse_as_neuro_image("reg.neu.ro/user/image:tag")
        assert str(image) == "image://default/user/image:tag"

    def test_parse_as_neuro_image_with_registry_prefix_special_chars(self) -> None:
        image = self.parser.parse_as_neuro_image(
            "reg.neu.ro/user/image%23%252d%3F%C3%9F:tag%23%252d%3F%C3%9F"
        )
        assert image == RemoteImage.new_neuro_image(
            name="image#%2d?ß",
            tag="tag#%2d?ß",
            owner="user",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )
        assert self.parser.parse_as_neuro_image(str(image)) == image

    def test_parse_as_neuro_image_no_scheme_3_slash_with_tag_fail(self) -> None:
        image = "something/docker.io/library/ubuntu:v10.04"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image)

    def test_parse_as_neuro_image_allow_tag_false_with_scheme_no_tag(self) -> None:
        image = "image:ubuntu"
        parsed = self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)
        assert parsed == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag=None,
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_neuro_image_allow_tag_false_no_scheme_no_tag(self) -> None:
        image = "ubuntu"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)

    def test_parse_as_neuro_image_allow_tag_false_no_scheme_with_tag(self) -> None:
        image = "ubuntu:latest"
        with pytest.raises(ValueError, match="tag is not allowed"):
            self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)

    def test_parse_as_neuro_image_allow_tag_false_with_scheme_lstrip(self) -> None:
        image = "image:game:latest"
        with pytest.raises(ValueError, match="tag is not allowed"):
            self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)
        image = "image:game:mega"
        with pytest.raises(ValueError, match="tag is not allowed"):
            self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)
        image = "image:game:v2.0:latest"
        with pytest.raises(ValueError, match="too many tags"):
            self.parser.parse_as_neuro_image(image, tag_option=TagOption.DENY)

    def test_convert_to_local_image(self) -> None:
        neuro_image = RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="artem",
            cluster_name="default",
            registry="reg.com",
            org_name=None,
        )
        local_image = self.parser.convert_to_local_image(neuro_image)
        assert local_image == LocalImage(name="ubuntu", tag="latest")

    def test_convert_to_neuro_image(self) -> None:
        local_image = LocalImage(name="ubuntu", tag="latest")
        neuro_image = self.parser.convert_to_neuro_image(local_image)
        assert neuro_image == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_convert_to_neuro_image_with_defualt_org(self) -> None:
        local_image = LocalImage(name="ubuntu", tag="latest")
        parser = _ImageNameParser(
            default_user="alice",
            default_cluster="default",
            default_org="test-org",
            registry_urls={
                "default": URL("https://reg.neu.ro"),
                "another": URL("https://example.org"),
            },
        )
        neuro_image = parser.convert_to_neuro_image(local_image)
        assert neuro_image == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="latest",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name="test-org",
        )

    def test_convert_to_neuro_image__neuro_registry(self) -> None:
        local_image = LocalImage(name="reg.neu.ro/bob/ubuntu", tag="v20.04")
        neuro_image = self.parser.convert_to_neuro_image(local_image)
        assert neuro_image == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v20.04",
            owner="bob",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_convert_to_neuro_image__neuro_registry__no_user(self) -> None:
        local_image = LocalImage(name="reg.neu.ro/ubuntu", tag="v20.04")
        neuro_image = self.parser.convert_to_neuro_image(local_image)
        assert neuro_image == RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v20.04",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_convert_to_neuro_image__neuro_registry__no_path(self) -> None:
        local_image = LocalImage(name="reg.neu.ro/", tag="v20.04")
        neuro_image = self.parser.convert_to_neuro_image(local_image)
        assert neuro_image == RemoteImage.new_neuro_image(
            name="reg.neu.ro/",
            tag="v20.04",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    # corner case 'image:latest'

    def test_parse_as_neuro_image__ambiguous_case__fail(self) -> None:
        url = "image:latest"
        with pytest.raises(ValueError, match="ambiguous value"):
            self.parser.parse_as_neuro_image(url)

    def test_parse_as_local_image__ambiguous_case__fail(self) -> None:
        url = "image:latest"
        with pytest.raises(ValueError, match="ambiguous value"):
            self.parser.parse_as_local_image(url)

    # other corner cases

    def test_parse_as_neuro_image__numeric_name(self) -> None:
        image = "image:5000"
        parsed = self.parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="5000",
            tag="latest",
            owner="alice",
            cluster_name="default",
            registry="reg.neu.ro",
            org_name=None,
        )

    def test_parse_as_local_image__neuro_registry(self) -> None:
        image = "reg.neu.ro/bob/ubuntu:v10.04"
        parsed = self.parser.parse_as_local_image(image)
        assert parsed == LocalImage(name="reg.neu.ro/bob/ubuntu", tag="v10.04")

    def test_parse_as_local_image__registry_has_port__neuro_registry(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "localhost:5000/bob/ubuntu:v10.04"
        parsed = my_parser.parse_as_local_image(image)
        assert parsed == LocalImage(name="localhost:5000/bob/ubuntu", tag="v10.04")

    def test_parse_as_neuro_image__registry_has_port__neuro_image(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "image://test-cluster/bob/library/ubuntu:v10.04"
        parsed = my_parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="test-cluster",
            registry="localhost:5000",
            org_name=None,
        )

    def test_parse_as_neuro_image__registry_has_port__image_in_good_repo(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "localhost:5000/bob/library/ubuntu:v10.04"
        parsed = my_parser.parse_as_neuro_image(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="test-cluster",
            registry="localhost:5000",
            org_name=None,
        )

    def test_parse_as_neuro_image__registry_has_port__image_in_bad_repo(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "localhost:9999/bob/library/ubuntu:v10.04"
        with pytest.raises(ValueError, match="scheme 'image:' is required"):
            my_parser.parse_as_neuro_image(image)

    def test_parse_remote__registry_has_port__neuro_image(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "image://test-cluster/bob/library/ubuntu:v10.04"
        parsed = my_parser.parse_remote(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="test-cluster",
            registry="localhost:5000",
            org_name=None,
        )

    def test_parse_remote__registry_has_port__image_in_good_repo(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "localhost:5000/bob/library/ubuntu:v10.04"
        parsed = my_parser.parse_remote(image)
        assert parsed == RemoteImage.new_neuro_image(
            name="library/ubuntu",
            tag="v10.04",
            owner="bob",
            cluster_name="test-cluster",
            registry="localhost:5000",
            org_name=None,
        )

    def test_parse_remote__registry_has_port__image_in_other_repo(self) -> None:
        my_parser = _ImageNameParser(
            default_user="alice",
            default_cluster="test-cluster",
            default_org=None,
            registry_urls={"test-cluster": URL("http://localhost:5000")},
        )
        image = "example.com:9999/bob/library/ubuntu:v10.04"
        parsed = my_parser.parse_remote(image)
        # NOTE: "owner" is parsed only for images in neuro registry
        assert parsed == RemoteImage.new_external_image(
            name="bob/library/ubuntu",
            tag="v10.04",
            registry="example.com:9999",
        )


class TestRemoteImage:
    def test_as_str_in_neuro_registry_tag_none(self) -> None:
        image = RemoteImage.new_neuro_image(
            name="ubuntu",
            tag=None,
            owner="me",
            cluster_name="test-cluster",
            org_name=None,
            registry="registry.io",
        )
        assert str(image) == "image://test-cluster/me/ubuntu"
        assert image.as_docker_url() == "registry.io/me/ubuntu"

    def test_as_str_in_neuro_registry_tag_yes(self) -> None:
        image = RemoteImage.new_neuro_image(
            name="ubuntu",
            tag="v10.04",
            owner="me",
            cluster_name="test-cluster",
            registry="registry.io",
            org_name=None,
        )
        assert str(image) == "image://test-cluster/me/ubuntu:v10.04"
        assert image.as_docker_url() == "registry.io/me/ubuntu:v10.04"

    def test_as_str_in_neuro_registry_tag_special_chars(self) -> None:
        image = RemoteImage.new_neuro_image(
            name="image#%2d?ß",
            tag="tag#%2d?ß",
            owner="me",
            cluster_name="test-cluster",
            registry="registry.io",
            org_name=None,
        )
        assert (
            str(image)
            == "image://test-cluster/me/image%23%252d%3F%C3%9F:tag%23%252d%3F%C3%9F"
        )
        assert image.as_docker_url() == "registry.io/me/image#%2d?ß:tag#%2d?ß"

    def test_as_str_not_in_neuro_registry_tag_none(self) -> None:
        image = RemoteImage.new_external_image(name="ubuntu")
        assert str(image) == "ubuntu"
        assert image.as_docker_url() == "ubuntu"

    def test_as_str_not_in_neuro_registry_tag_yes(self) -> None:
        image = RemoteImage.new_external_image(name="ubuntu", tag="v10.04")
        assert str(image) == "ubuntu:v10.04"
        assert image.as_docker_url() == "ubuntu:v10.04"

    def test_as_docker_url_in_neuro_registry(self) -> None:
        image = RemoteImage(
            name="ubuntu",
            tag="v10.04",
            owner="me",
            cluster_name="test-cluster",
            registry="registry.io",
        )
        assert image.as_docker_url() == "registry.io/me/ubuntu:v10.04"

    def test_as_docker_url_not_in_neuro_registry(self) -> None:
        image = RemoteImage(name="ubuntu", tag="v10.04", owner=None, registry=None)
        assert image.as_docker_url() == "ubuntu:v10.04"


@pytest.mark.usefixtures("patch_docker_host")
class TestImages:
    parser = _ImageNameParser(
        default_user="user",
        default_cluster="default",
        default_org=None,
        registry_urls={"default": URL("https://registry-dev.neu.ro")},
    )

    @mock.patch(
        "aiodocker.Docker.__init__",
        side_effect=ValueError(
            "text Either DOCKER_HOST or local sockets are not available text"
        ),
    )
    async def test_unavailable_docker(
        self, patched_init: Any, make_client: _MakeClient
    ) -> None:
        image = self.parser.parse_as_neuro_image("image://default/bob/image:bananas")
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(DockerError, match=r"Docker engine is not available.+"):
                await client.images.pull(image, local_image)

    @mock.patch(
        "aiodocker.Docker.__init__", side_effect=ValueError("something went wrong")
    )
    async def test_unknown_docker_error(
        self, patched_init: Any, make_client: _MakeClient
    ) -> None:
        image = self.parser.parse_as_neuro_image("image://default/bob/image:bananas")
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(ValueError, match=r"something went wrong"):
                await client.images.pull(image, local_image)

    @mock.patch("aiodocker.images.DockerImages.tag")
    async def test_push_non_existent_image(
        self, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        patched_tag.side_effect = DockerError(404, {"message": "Mocked error"})
        image = self.parser.parse_as_neuro_image(
            "image://default/bob/image:bananas-no-more"
        )
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(ValueError, match=r"not found"):
                await client.images.push(local_image, image)

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.push")
    async def test_push_image_to_foreign_repo(
        self, patched_push: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        patched_tag.return_value = True
        patched_push.side_effect = DockerError(403, {"message": "Mocked error"})
        image = self.parser.parse_as_neuro_image(
            "image://default/bob/image:bananas-no-more"
        )
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(AuthorizationError):
                await client.images.push(local_image, image)

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.push")
    async def test_push_image_with_docker_api_error(
        self, patched_push: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        async def error_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {"error": True, "errorDetail": {"message": "Mocked message"}}

        patched_tag.return_value = True
        patched_push.return_value = error_generator()
        image = self.parser.parse_as_neuro_image(
            "image://default/bob/image:bananas-wrong-food"
        )
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(DockerError) as exc_info:
                await client.images.push(local_image, image)
        assert exc_info.value.status == 900
        assert exc_info.value.message == "Mocked message"

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.push")
    async def test_success_push_image(
        self, patched_push: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        async def message_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {}

        patched_tag.return_value = True
        patched_push.return_value = message_generator()
        image = self.parser.parse_as_neuro_image(
            "image://default/bob/image:bananas-is-here"
        )
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            result = await client.images.push(local_image, image)
        assert result == image

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.push")
    async def test_success_push_image_no_target(
        self, patched_push: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        async def message_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {}

        patched_tag.return_value = True
        patched_push.return_value = message_generator()
        image = self.parser.parse_as_neuro_image("image://default/user/bananas:latest")
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            result = await client.images.push(local_image)
        assert result == image

    @mock.patch("aiodocker.images.DockerImages.pull")
    async def test_pull_non_existent_image(
        self, patched_pull: Any, make_client: _MakeClient
    ) -> None:
        patched_pull.side_effect = DockerError(404, {"message": "Mocked error"})
        async with make_client("https://api.localhost.localdomain") as client:
            image = self.parser.parse_as_neuro_image(
                "image://default/bob/image:no-bananas-here"
            )
            local_image = self.parser.parse_as_local_image("bananas:latest")
            with pytest.raises(ValueError, match=r"not found"):
                await client.images.pull(image, local_image)

    @mock.patch("aiodocker.images.DockerImages.pull")
    async def test_pull_image_from_foreign_repo(
        self, patched_pull: Any, make_client: _MakeClient
    ) -> None:
        patched_pull.side_effect = DockerError(403, {"message": "Mocked error"})
        image = self.parser.parse_as_neuro_image(
            "image://default/bob/image:not-your-bananas"
        )
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(AuthorizationError):
                await client.images.pull(image, local_image)

    @mock.patch("aiodocker.images.DockerImages.pull")
    async def test_pull_image_with_docker_api_error(
        self, patched_pull: Any, make_client: Any
    ) -> None:
        async def error_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {"error": True, "errorDetail": {"message": "Mocked message"}}

        patched_pull.return_value = error_generator()
        image = self.parser.parse_as_neuro_image("image://default/bob/image:nuts-here")
        async with make_client("https://api.localhost.localdomain") as client:
            with pytest.raises(DockerError) as exc_info:
                await client.images.pull(image, image)
        assert exc_info.value.status == 900
        assert exc_info.value.message == "Mocked message"

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.pull")
    async def test_success_pull_image(
        self, patched_pull: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        async def message_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {}

        patched_tag.return_value = True
        patched_pull.return_value = message_generator()
        image = self.parser.parse_as_neuro_image("image://default/bob/image:bananas")
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            result = await client.images.pull(image, local_image)
        assert result == local_image

    @mock.patch("aiodocker.images.DockerImages.tag")
    @mock.patch("aiodocker.images.DockerImages.pull")
    async def test_success_pull_image_no_target(
        self, patched_pull: Any, patched_tag: Any, make_client: _MakeClient
    ) -> None:
        async def message_generator() -> AsyncIterator[Dict[str, Any]]:
            yield {}

        patched_tag.return_value = True
        patched_pull.return_value = message_generator()
        image = self.parser.parse_as_neuro_image("image://default/bob/bananas:latest")
        local_image = self.parser.parse_as_local_image("bananas:latest")
        async with make_client("https://api.localhost.localdomain") as client:
            result = await client.images.pull(image)
        assert result == local_image


class TestRegistry:
    @pytest.mark.skipif(
        sys.platform == "win32", reason="aiodocker doesn't support Windows pipes yet"
    )
    async def test_ls_repositories(
        self, aiohttp_server: _TestServerFactory, make_client: _MakeClient
    ) -> None:
        JSON = {"repositories": ["bob/alpine", "jill/bananas"]}

        async def handler(request: web.Request) -> web.Response:
            assert "n" in request.query
            assert "last" not in request.query
            return web.json_response(JSON)

        app = web.Application()
        app.router.add_get("/v2/_catalog", handler)

        srv = await aiohttp_server(app)
        url = "http://platform"
        registry_url = srv.make_url("/v2/")

        async with make_client(url, registry_url=registry_url) as client:
            ret = await client.images.list()

        registry = _get_url_authority(registry_url)
        assert registry is not None
        assert set(ret) == {
            RemoteImage.new_neuro_image(
                "alpine",
                tag=None,
                owner="bob",
                cluster_name="default",
                registry=registry,
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "bananas",
                tag=None,
                owner="jill",
                cluster_name="default",
                registry=registry,
                org_name=None,
            ),
        }

    async def test_ls_repositories_chunked(
        self, aiohttp_server: _TestServerFactory, make_client: _MakeClient
    ) -> None:
        step = 0

        async def handler(request: web.Request) -> web.Response:
            nonlocal step
            step += 1
            headers: Dict[str, str]
            assert "n" in request.query
            if step == 1:
                assert "last" not in request.query
                payload = {"repositories": ["bob/alpine", "jill/bananas"]}
                headers = {LINK: f'<{catalog_url}?last=lsttkn>; rel="next"'}
                return web.json_response(payload, headers=headers)
            elif step == 2:
                assert request.query["last"] == "lsttkn"
                payload = {"repositories": ["alice/library/ubuntu"]}
                headers = {LINK: f'<{catalog_url}?last=lsttkn2>; rel="next"'}
                return web.json_response(payload, headers=headers)
            elif step == 3:
                assert request.query["last"] == "lsttkn2"
                payload = {"repositories": []}
                headers = {LINK: f'<{catalog_url}?last=lsttkn3>; rel="next"'}
                return web.json_response(payload, headers=headers)
            else:  # pragma: no cover
                assert False

        app = web.Application()
        app.router.add_get("/v2/_catalog", handler)

        srv = await aiohttp_server(app)
        url = "http://platform"
        registry_url = srv.make_url("/v2/")
        catalog_url = registry_url / "_catalog"

        registry = _get_url_authority(registry_url)
        assert registry is not None
        expected = {
            RemoteImage.new_neuro_image(
                "alpine",
                tag=None,
                owner="bob",
                cluster_name="default",
                registry=registry,
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "library/ubuntu",
                tag=None,
                owner="alice",
                cluster_name="default",
                registry=registry,
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "bananas",
                tag=None,
                owner="jill",
                cluster_name="default",
                registry=registry,
                org_name=None,
            ),
        }

        async with make_client(url, registry_url=registry_url) as client:
            ret = await client.images.list()
        assert step == 3  # All steps are passed
        assert set(ret) == expected

    @pytest.mark.skipif(
        sys.platform == "win32", reason="aiodocker doesn't support Windows pipes yet"
    )
    async def test_ls_tags(
        self, aiohttp_server: _TestServerFactory, make_client: _MakeClient
    ) -> None:
        JSON = {"name": "test", "tags": ["alpha", "beta", "gamma"]}

        async def handler(request: web.Request) -> web.Response:
            assert "n" in request.query
            assert "last" not in request.query
            return web.json_response(JSON)

        app = web.Application()
        app.router.add_get("/v2/me/test/tags/list", handler)

        srv = await aiohttp_server(app)
        url = "http://platform"
        registry_url = srv.make_url("/v2/")

        async with make_client(url, registry_url=registry_url) as client:
            image = RemoteImage.new_neuro_image(
                name="test",
                tag=None,
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            )
            ret = await client.images.tags(image)

        assert set(ret) == {
            RemoteImage.new_neuro_image(
                "test",
                tag="alpha",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "test",
                tag="beta",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "test",
                tag="gamma",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
        }

    async def test_ls_tags_chunked(
        self, aiohttp_server: _TestServerFactory, make_client: _MakeClient
    ) -> None:
        step = 0

        async def handler(request: web.Request) -> web.Response:
            nonlocal step
            step += 1
            headers: Dict[str, str]
            assert "n" in request.query
            if step == 1:
                assert "last" not in request.query
                payload = {"name": "test", "tags": ["alpha", "beta"]}
                headers = {LINK: f'<{tags_list_url}?last=lsttkn>; rel="next"'}
                return web.json_response(payload, headers=headers)
            elif step == 2:
                assert request.query["last"] == "lsttkn"
                payload = {"name": "test", "tags": ["gamma"]}
                headers = {LINK: f'<{tags_list_url}?last=lsttkn2>; rel="next"'}
                return web.json_response(payload, headers=headers)
            elif step == 3:
                assert request.query["last"] == "lsttkn2"
                payload = {"name": "test", "tags": []}
                headers = {LINK: f'<{tags_list_url}?last=lsttkn3>; rel="next"'}
                return web.json_response(payload, headers=headers)
            else:  # pragma: no cover
                assert False

        app = web.Application()
        app.router.add_get("/v2/me/test/tags/list", handler)

        srv = await aiohttp_server(app)
        url = "http://platform"
        registry_url = srv.make_url("/v2/")
        tags_list_url = registry_url / "me/test/tags/list"

        async with make_client(url, registry_url=registry_url) as client:
            image = RemoteImage.new_neuro_image(
                name="test",
                tag=None,
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            )
            ret = await client.images.tags(image)
        assert step == 3  # All steps are passed

        assert set(ret) == {
            RemoteImage.new_neuro_image(
                "test",
                tag="alpha",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "test",
                tag="beta",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
            RemoteImage.new_neuro_image(
                "test",
                tag="gamma",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            ),
        }

    @pytest.mark.skipif(
        sys.platform == "win32", reason="aiodocker doesn't support Windows pipes yet"
    )
    async def test_tag_info(
        self, aiohttp_server: _TestServerFactory, make_client: _MakeClient
    ) -> None:
        JSON = {
            "schemaVersion": 2,
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "config": {
                "mediaType": "application/vnd.docker.container.image.v1+json",
                "size": 349,
                "digest": "sha256:ignored-1",
            },
            "layers": [
                {
                    "mediaType": "application/vnd.docker.image.rootfs.diff.tar.gzip",
                    "size": 32036078,
                    "digest": "sha256:ignored-2",
                },
                {
                    "mediaType": "application/vnd.docker.image.rootfs.diff.tar.gzip",
                    "size": 854454,
                    "digest": "sha256:ignored-3",
                },
            ],
        }

        async def handler(request: web.Request) -> web.Response:
            assert (
                request.headers["Accept"]
                == "application/vnd.docker.distribution.manifest.v2+json"
            )
            return web.json_response(JSON)

        app = web.Application()
        app.router.add_get("/v2/me/test/manifests/test_tag", handler)

        srv = await aiohttp_server(app)
        url = "http://platform"
        registry_url = srv.make_url("/v2/")

        async with make_client(url, registry_url=registry_url) as client:
            image = RemoteImage.new_neuro_image(
                name="test",
                tag="test_tag",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            )
            ret = await client.images.tag_info(image)
            assert image.tag

        assert ret == Tag(name=image.tag, size=32890532)

    @pytest.mark.skipif(
        sys.platform == "win32", reason="aiodocker doesn't support Windows pipes yet"
    )
    async def test_tags_bad_image_with_tag(self, make_client: _MakeClient) -> None:
        url = URL("http://whatever")
        registry_url = URL("http://whatever-registry")
        async with make_client(url, registry_url=registry_url) as client:
            image = RemoteImage.new_neuro_image(
                name="ubuntu",
                tag="latest",
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            )
            with pytest.raises(ValueError, match="tag is not allowed"):
                await client.images.tags(image)

    @pytest.mark.skipif(
        sys.platform == "win32", reason="aiodocker doesn't support Windows pipes yet"
    )
    async def test_tags_bad_image_without_name(self, make_client: _MakeClient) -> None:
        url = URL("http://whatever")
        registry_url = URL("http://whatever-registry")
        async with make_client(url, registry_url=registry_url) as client:
            image = RemoteImage.new_neuro_image(
                name="",
                tag=None,
                owner="me",
                cluster_name="default",
                registry="reg",
                org_name=None,
            )
            with pytest.raises(ValueError, match="missing image name"):
                await client.images.tags(image)



import base64
from dataclasses import dataclass
from typing import AsyncIterator, Optional

from yarl import URL

from ._config import Config
from ._core import _Core
from ._rewrite import rewrite_module
from ._utils import NoPublicConstructor, asyncgeneratorcontextmanager


@rewrite_module
@dataclass(frozen=True)
class Secret:
    key: str
    owner: str
    cluster_name: str
    org_name: Optional[str]

    @property
    def uri(self) -> URL:
        base = f"secret://{self.cluster_name}"
        if self.org_name:
            base += f"/{self.org_name}"
        return URL(f"{base}/{self.owner}/{self.key}")


@rewrite_module
class Secrets(metaclass=NoPublicConstructor):
    def __init__(self, core: _Core, config: Config) -> None:
        self._core = core
        self._config = config

    def _get_secrets_url(self, cluster_name: Optional[str]) -> URL:
        if cluster_name is None:
            cluster_name = self._config.cluster_name
        return self._config.get_cluster(cluster_name).secrets_url

    @asyncgeneratorcontextmanager
    async def list(self, cluster_name: Optional[str] = None) -> AsyncIterator[Secret]:
        url = self._get_secrets_url(cluster_name)
        auth = await self._config._api_auth()
        async with self._core.request("GET", url, auth=auth) as resp:
            ret = await resp.json()
            for j in ret:
                yield Secret(
                    key=j["key"],
                    owner=j["owner"],
                    cluster_name=cluster_name or self._config.cluster_name,
                    org_name=j.get("org_name"),
                )

    async def add(
        self,
        key: str,
        value: bytes,
        cluster_name: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> None:
        url = self._get_secrets_url(cluster_name)
        auth = await self._config._api_auth()
        data = {
            "key": key,
            "value": base64.b64encode(value).decode("ascii"),
            "org_name": org_name or self._config.org_name,
        }
        async with self._core.request("POST", url, auth=auth, json=data):
            pass

    async def rm(
        self,
        key: str,
        cluster_name: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> None:
        url = self._get_secrets_url(cluster_name) / key
        auth = await self._config._api_auth()
        params = {}
        org_name = org_name or self._config.org_name
        if org_name:
            params["org_name"] = org_name
        async with self._core.request("DELETE", url, auth=auth, params=params):
            pass



import asyncio
import sys
from pathlib import Path
from typing import Any, AsyncIterator, Callable

import pytest
from yarl import URL

from neuro_sdk import Client
from neuro_sdk._url_utils import (
    _extract_path,
    normalize_local_path_uri,
    normalize_storage_path_uri,
    uri_from_cli,
)


@pytest.fixture
async def client(
    loop: asyncio.AbstractEventLoop, make_client: Callable[..., Client]
) -> AsyncIterator[Client]:
    async with make_client("https://example.com") as client:
        yield client


# asvetlov: I don't like autouse but it is the fastest fix
@pytest.fixture(autouse=True)
def fake_homedir(monkeypatch: Any, tmp_path: Path) -> Path:
    monkeypatch.setenv("HOME", str(tmp_path))
    return Path.home()


@pytest.fixture
def pwd() -> Path:
    return Path.cwd()


async def test_config_username(token: str, client: Client) -> None:
    assert client.username == "user"


def test_uri_from_cli_relative_path() -> None:
    uri = uri_from_cli("path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == Path("path/to/file.txt").absolute().as_uri()


def test_uri_from_cli_absolute_path() -> None:
    uri = uri_from_cli("/path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == Path("/path/to/file.txt").absolute().as_uri()


def test_uri_from_cli_relative_path_special_chars() -> None:
    uri = uri_from_cli("path/to/file#%23:?@~", "testuser", "test-cluster", None)
    assert uri.path.endswith("/path/to/file#%23:?@~")


def test_uri_from_cli_absolute_path_special_chars() -> None:
    uri = uri_from_cli("/path/to/file#%23:?@~", "testuser", "test-cluster", None)
    assert _extract_path(uri) == Path("/path/to/file#%23:?@~").absolute()


def test_uri_from_cli_path_with_tilde(fake_homedir: Path) -> None:
    uri = uri_from_cli("~/path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == (fake_homedir / "path/to/file.txt").as_uri()


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="expanduser() does not fail for unknown user on Windows",
)
def test_uri_from_cli_path_with_tilde_unknown_user() -> None:
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        uri_from_cli("~unknownuser/path/to/file.txt", "testuser", "test-cluster", None)


def test_uri_from_cli_tilde_only(fake_homedir: Path) -> None:
    uri = uri_from_cli("~", "testuser", "test-cluster", None)
    assert str(uri) == fake_homedir.as_uri()


def test_uri_from_cli_relative_file_uri() -> None:
    uri = uri_from_cli("file:path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == Path("path/to/file.txt").absolute().as_uri()


def test_uri_from_cli_absolute_file_uri() -> None:
    uri = uri_from_cli("file:/path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == Path("/path/to/file.txt").absolute().as_uri()
    uri = uri_from_cli("file:///path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == Path("/path/to/file.txt").absolute().as_uri()


def test_uri_from_cli_relative_file_uri_special_chars() -> None:
    uri = uri_from_cli(
        "file:path/to/file%23%252d%3f:@~%C3%9F", "testuser", "test-cluster", None
    )
    assert uri.path.endswith("/path/to/file#%2d?:@~ß")


def test_uri_from_cli_absolute_file_uri_special_chars() -> None:
    uri = uri_from_cli(
        "file:/path/to/file%23%252d%3f:@~%C3%9F", "testuser", "test-cluster", None
    )
    assert uri.path.endswith("/path/to/file#%2d?:@~ß")


def test_uri_from_cli_relative_storage_uri() -> None:
    uri = uri_from_cli("storage:path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == "storage://test-cluster/testuser/path/to/file.txt"
    uri = uri_from_cli("storage:/path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == "storage://test-cluster/path/to/file.txt"


def test_uri_from_cli_absolute_storage_uri() -> None:
    uri = uri_from_cli(
        "storage://otheruser/path/to/file.txt", "testuser", "test-cluster", None
    )
    assert str(uri) == "storage://otheruser/path/to/file.txt"
    uri = uri_from_cli("storage:///path/to/file.txt", "testuser", "test-cluster", None)
    assert str(uri) == "storage://test-cluster/path/to/file.txt"


def test_uri_from_cli_absolute_storage_uri_special_chars() -> None:
    uri = uri_from_cli(
        "storage://cluster/user/path/to/file%23%252d%3f:@~%C3%9F",
        "testuser",
        "test-cluster",
        None,
    )
    assert uri.path == "/user/path/to/file#%2d?:@~ß"


def test_uri_from_cli_numberic_path() -> None:
    uri = uri_from_cli("256", "testuser", "test-cluster", None)
    assert str(uri) == Path("256").absolute().as_uri()
    uri = uri_from_cli("123456", "testuser", "test-cluster", None)
    assert str(uri) == Path("123456").absolute().as_uri()
    uri = uri_from_cli("file:256", "testuser", "test-cluster", None)
    assert str(uri) == Path("256").absolute().as_uri()
    uri = uri_from_cli("file:123456", "testuser", "test-cluster", None)
    assert str(uri) == Path("123456").absolute().as_uri()
    uri = uri_from_cli("storage:256", "testuser", "test-cluster", None)
    assert str(uri) == "storage://test-cluster/testuser/256"
    uri = uri_from_cli("storage:123456", "testuser", "test-cluster", None)
    assert str(uri) == "storage://test-cluster/testuser/123456"


@pytest.mark.parametrize(
    "path_or_uri",
    [
        "https://cluster/user/path/to",
        "file://cluster/user/path/to",
        "file:/path/to#fragment",
        "file:/path/to#",
        "file:/path/to?key=value",
        "file:/path/to?",
    ],
)
async def test_uri_from_cli__file__fail(path_or_uri: str) -> None:
    with pytest.raises(ValueError):
        uri_from_cli(path_or_uri, "u", "c", None, allowed_schemes=("file",))


@pytest.mark.parametrize(
    "uri",
    [
        "",
        "https://cluster/user/path/to",
        "storage://cluster/user/path/to#fragment",
        "storage://cluster/user/path/to#",
        "storage://cluster/user/path/to?key=value",
        "storage://cluster/user/path/to?",
        "storage://user@cluster/user/path/to",
        "storage://:password@cluster/user/path/to",
        "storage://:@cluster/user/path/to",
        "storage://cluster:1234/user/path/to",
    ],
)
async def test_uri_from_cli__storage__fail(uri: str) -> None:
    with pytest.raises(ValueError):
        uri_from_cli(uri, "u", "c", None, allowed_schemes=("storage",))


@pytest.mark.parametrize(
    "uri",
    [
        "",
        "https://cluster/user/image",
        "image://cluster/user/image#fragment",
        "image://cluster/user/image#",
        "image://cluster/user/image?key=value",
        "image://cluster/user/image?",
        "image://user@cluster/user/image",
        "image://:password@cluster/user/image",
        "image://:@cluster/user/image",
        "image://cluster:1234/user/image",
    ],
)
async def test_uri_from_cli__image__fail(uri: str) -> None:
    with pytest.raises(ValueError):
        uri_from_cli(uri, "u", "c", None, allowed_schemes=("image",))


@pytest.mark.parametrize(
    "uri",
    [
        "",
        "https://cluster/bucket/object",
        "blob://cluster/bucket/object#fragment",
        "blob://cluster/bucket/object#",
        "blob://cluster/bucket/object?key=value",
        "blob://cluster/bucket/object?",
        "blob://user@cluster/bucket/object",
        "blob://:password@cluster/bucket/object",
        "blob://:@cluster/bucket/object",
        "blob://cluster:1234/bucket/object",
    ],
)
async def test_uri_from_cli__blob__fail(uri: str) -> None:
    with pytest.raises(ValueError):
        uri_from_cli(uri, "u", "c", None, allowed_schemes=("blob",))


async def test_normalize_storage_path_uri_no_path(client: Client) -> None:
    url = URL("storage:")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user"
    assert str(url) == "storage://test-cluster/user"


async def test_normalize_local_path_uri_no_path(pwd: Path) -> None:
    url = URL("file:")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd


async def test_normalize_storage_path_uri_no_slashes(client: Client) -> None:
    url = URL("storage:file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user/file.txt"
    assert str(url) == "storage://test-cluster/user/file.txt"


async def test_normalize_storage_path_uri_no_slashes_with_org(client: Client) -> None:
    url = URL("storage:file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", "test-org")
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/test-org/user/file.txt"
    assert str(url) == "storage://test-cluster/test-org/user/file.txt"


async def test_normalize_local_path_uri_no_slashes(pwd: Path) -> None:
    url = URL("file:file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "file.txt"


async def test_normalize_storage_path_uri__0_slashes_relative(client: Client) -> None:
    url = URL("storage:path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user/path/to/file.txt"
    assert str(url) == "storage://test-cluster/user/path/to/file.txt"


async def test_normalize_storage_path_uri__0_slashes_relative_with_org(
    client: Client,
) -> None:
    url = URL("storage:path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", "test-org")
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/test-org/user/path/to/file.txt"
    assert str(url) == "storage://test-cluster/test-org/user/path/to/file.txt"


async def test_normalize_local_path_uri__0_slashes_relative(pwd: Path) -> None:
    url = URL("file:path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "path/to/file.txt"


async def test_normalize_storage_path_uri__1_slash_absolute(client: Client) -> None:
    url = URL("storage:/path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/path/to/file.txt"
    assert str(url) == "storage://test-cluster/path/to/file.txt"


async def test_normalize_storage_path_uri__1_slash_absolute_with_org(
    client: Client,
) -> None:
    url = URL("storage:/path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", "test-org")
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/test-org/path/to/file.txt"
    assert str(url) == "storage://test-cluster/test-org/path/to/file.txt"


async def test_normalize_local_path_uri__1_slash_absolute(pwd: Path) -> None:
    url = URL("file:/path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == Path(pwd.drive + "/path/to/file.txt")


async def test_normalize_storage_path_uri__2_slashes(client: Client) -> None:
    url = URL("storage://path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "path"
    assert url.path == "/to/file.txt"
    assert str(url) == "storage://path/to/file.txt"


async def test_normalize_local_path_uri__2_slashes(pwd: Path) -> None:
    url = URL("file://path/to/file.txt")
    with pytest.raises(
        ValueError, match="Host part is not allowed in file URI, found 'path'"
    ):
        url = normalize_local_path_uri(url)


async def test_normalize_storage_path_uri__3_slashes_relative(client: Client) -> None:
    url = URL("storage:///path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/path/to/file.txt"
    assert str(url) == "storage://test-cluster/path/to/file.txt"


async def test_normalize_local_path_uri__3_slashes_relative(pwd: Path) -> None:
    url = URL("file:///path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == Path(pwd.drive + "/path/to/file.txt")


async def test_normalize_storage_path_uri__4_slashes_relative(client: Client) -> None:
    url = URL("storage:////path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/path/to/file.txt"
    assert str(url) == "storage://test-cluster/path/to/file.txt"


@pytest.mark.skipif(sys.platform == "win32", reason="Doesn't work on Windows")
async def test_normalize_local_path_uri__4_slashes_relative() -> None:
    url = URL("file:////path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert url.path == "/path/to/file.txt"
    assert str(url) == f"file:///path/to/file.txt"


@pytest.mark.parametrize(
    "uri_str",
    [
        "",
        "https://cluster/user/path/to",
        "storage://cluster/user/path/to#fragment",
        "storage://cluster/user/path/to?key=value",
        "storage://user@cluster/user/path/to",
        "storage://:password@cluster/user/path/to",
        "storage://:@cluster/user/path/to",
        "storage://cluster:1234/user/path/to",
    ],
)
async def test_normalize_storage_path_uri__fail(uri_str: str) -> None:
    uri = URL(uri_str)
    with pytest.raises(ValueError):
        normalize_storage_path_uri(uri, "test-user", "test-cluster", None)


async def test_normalize_storage_path_uri__tilde_in_relative_path(
    client: Client,
) -> None:
    url = URL("storage:~/path/to/file.txt")
    with pytest.raises(ValueError, match=".*Cannot expand user.*"):
        normalize_storage_path_uri(url, client.username, "test-cluster", None)


async def test_normalize_local_path_uri__tilde_in_relative_path(
    fake_homedir: Path,
) -> None:
    url = URL("file:~/path/to/file.txt")
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        normalize_local_path_uri(url)


async def test_normalize_storage_path_uri__tilde_in_relative_path_2(
    client: Client,
) -> None:
    url = URL("storage:./~/path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user/~/path/to/file.txt"
    assert str(url) == "storage://test-cluster/user/~/path/to/file.txt"


async def test_normalize_local_path_uri__tilde_in_relative_path_2(
    pwd: Path,
) -> None:
    url = URL("file:./~/path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "~/path/to/file.txt"
    assert str(url) == (pwd / "~/path/to/file.txt").as_uri().replace("%7E", "~")


async def test_normalize_storage_path_uri__tilde_in_relative_path_3(
    client: Client,
) -> None:
    url = URL("storage:path/to~file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user/path/to~file.txt"
    assert str(url) == "storage://test-cluster/user/path/to~file.txt"


async def test_normalize_local_path_uri__tilde_in_relative_path_3(
    fake_homedir: Path, pwd: Path
) -> None:
    url = URL("file:path/to~file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "path/to~file.txt"
    assert str(url) == (pwd / "path/to~file.txt").as_uri().replace("%7E", "~")


async def test_normalize_storage_path_uri__tilde_in_absolute_path(
    client: Client,
) -> None:
    url = URL("storage:/~/path/to/file.txt")
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        normalize_storage_path_uri(url, client.username, "test-cluster", None)


async def test_normalize_local_path_uri__tilde_in_absolute_path(
    fake_homedir: Path, pwd: Path
) -> None:
    url = URL("file:/~/path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "/~/path/to/file.txt"
    assert str(url) == (pwd / "/~/path/to/file.txt").as_uri().replace("%7E", "~")


async def test_normalize_storage_path_uri__tilde_in_host(client: Client) -> None:
    url = URL("storage://~/path/to/file.txt")
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        normalize_storage_path_uri(url, client.username, "test-cluster", None)


async def test_normalize_local_path_uri__tilde_in_host(
    client: Client, pwd: Path
) -> None:
    url = URL("file://~/path/to/file.txt")
    with pytest.raises(
        ValueError, match=f"Host part is not allowed in file URI, found '~'"
    ):
        url = normalize_local_path_uri(url)


async def test_normalize_storage_path_uri__bad_scheme(client: Client) -> None:
    with pytest.raises(ValueError, match="Invalid storage scheme 'other:'"):
        url = URL("other:path/to/file.txt")
        normalize_storage_path_uri(url, client.username, "test-cluster", None)


async def test_normalize_local_path_uri__bad_scheme() -> None:
    with pytest.raises(ValueError, match="Invalid local file scheme 'other:'"):
        url = URL("other:path/to/file.txt")
        normalize_local_path_uri(url)


# The tests below check that f(f(x)) == f(x) where f is a path normalization function


async def test_normalize_storage_path_uri__no_slash__double(client: Client) -> None:
    url = URL("storage:path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/user/path/to/file.txt"
    assert str(url) == "storage://test-cluster/user/path/to/file.txt"


async def test_normalize_local_path_uri__no_slash__double(pwd: Path) -> None:
    url = URL("file:path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "path/to/file.txt"


async def test_normalize_storage_path_uri__tilde_slash__double(client: Client) -> None:
    url = URL("storage:~/path/to/file.txt")
    with pytest.raises(ValueError, match=".*Cannot expand user.*"):
        normalize_storage_path_uri(url, client.username, "test-cluster", None)


async def test_normalize_local_path_uri__tilde_slash__double() -> None:
    url = URL("file:~/path/to/file.txt")
    with pytest.raises(ValueError, match=r"Cannot expand user for "):
        normalize_local_path_uri(url)


async def test_normalize_storage_path_uri__3_slashes__double(client: Client) -> None:
    url = URL("storage:///path/to/file.txt")
    url = normalize_storage_path_uri(url, client.username, "test-cluster", None)
    assert url.scheme == "storage"
    assert url.host == "test-cluster"
    assert url.path == "/path/to/file.txt"
    assert str(url) == "storage://test-cluster/path/to/file.txt"


async def test_normalize_local_path_uri__3_slashes__double(pwd: Path) -> None:
    url = URL(f"file:///{pwd}/path/to/file.txt")
    url = normalize_local_path_uri(url)
    assert url.scheme == "file"
    assert url.host is None
    assert _extract_path(url) == pwd / "path/to/file.txt"
    assert str(url) == (pwd / "path/to/file.txt").as_uri()


@pytest.mark.skipif(sys.platform != "win32", reason="Requires Windows")
def test_normalized_path() -> None:
    p = URL("file:///Z:/neuro/platform-api-clients/python/setup.py")
    assert normalize_local_path_uri(p) == p


@pytest.mark.parametrize(
    "uri_str",
    [
        "",
        "https://cluster/user/path/to",
        "file://cluster/user/path/to",
        "file:/path/to#fragment",
        "file:/path/to?key=value",
    ],
)
async def test_normalize_local_path_uri__fail(uri_str: str) -> None:
    uri = URL(uri_str)
    with pytest.raises(ValueError):
        normalize_local_path_uri(uri)



import contextlib
import errno
import json as jsonmodule
import logging
import sqlite3
import time
from contextlib import asynccontextmanager
from http.cookies import Morsel, SimpleCookie
from types import SimpleNamespace
from typing import (
    Any,
    AsyncIterator,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import aiohttp
from aiohttp import WSMessage
from multidict import CIMultiDict
from yarl import URL

from ._errors import (
    AuthenticationError,
    AuthorizationError,
    ClientError,
    IllegalArgumentError,
    ResourceNotFound,
    ServerNotAvailable,
)
from ._tracing import gen_trace_id
from ._utils import asyncgeneratorcontextmanager

log = logging.getLogger(__package__)

SESSION_COOKIE_MAXAGE = 5 * 60  # 5 min

SCHEMA = {
    "cookie_session": (
        "CREATE TABLE cookie_session "
        "(name TEXT, domain TEXT, path TEXT, cookie TEXT, timestamp REAL)"
    ),
    "cookie_session_index": (
        "CREATE UNIQUE INDEX cookie_session_index ON cookie_session " "(name)"
    ),
}
DROP = {
    "cookie_session_index": "DROP INDEX IF EXISTS cookie_session_index",
    "cookie_session": "DROP TABLE IF EXISTS cookie_session",
}


DEFAULT_TIMEOUT = aiohttp.ClientTimeout(None, None, 60, 60)


class _Core:
    """Transport provider for public API client.

    Internal class.
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        trace_id: Optional[str],
        trace_sampled: Optional[bool] = None,
    ) -> None:
        self._session = session
        self._trace_id = trace_id
        self._trace_sampled = trace_sampled
        self._exception_map = {
            400: IllegalArgumentError,
            401: AuthenticationError,
            403: AuthorizationError,
            404: ResourceNotFound,
            405: ClientError,
            502: ServerNotAvailable,
        }
        self._prev_cookie: Optional[Morsel[str]] = None

    def _post_init(
        self,
        db: sqlite3.Connection,
    ) -> None:
        for cookie in _load_cookies(db):
            self._session.cookie_jar.update_cookies({cookie.key: cookie})

    def _save_cookies(self, db: sqlite3.Connection) -> None:
        to_save = []
        for cookie in self._session.cookie_jar:
            name = cookie.key
            if name.startswith("NEURO_") and name.endswith("_SESSION"):
                to_save.append(cookie)
        _save_cookies(db, to_save)

    @property
    def timeout(self) -> aiohttp.ClientTimeout:
        # TODO: implement ClientSession.timeout public property for session
        return self._session._timeout

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    async def close(self) -> None:
        pass

    @asynccontextmanager
    async def request(
        self,
        method: str,
        url: URL,
        *,
        auth: str,
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
        data: Any = None,
        json: Any = None,
        headers: Optional[Mapping[str, str]] = None,
        timeout: Optional[aiohttp.ClientTimeout] = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        assert url.is_absolute()
        if params:
            url = url.with_query(params)
        log.debug("Fetch [%s] %s", method, url)
        if headers is not None:
            real_headers: CIMultiDict[str] = CIMultiDict(headers)
        else:
            real_headers = CIMultiDict()
        if len(auth.split()) > 1:
            # auth contains scheme and parameter
            real_headers["Authorization"] = auth
        if "Content-Type" not in real_headers:
            if json is not None:
                real_headers["Content-Type"] = "application/json"
        trace_request_ctx = SimpleNamespace()
        trace_id = self._trace_id
        if trace_id is None:
            trace_id = gen_trace_id()
        trace_request_ctx.trace_id = trace_id
        trace_request_ctx.trace_sampled = self._trace_sampled
        async with self._session.request(
            method,
            url,
            headers=real_headers,
            json=json,
            data=data,
            timeout=timeout,
            trace_request_ctx=trace_request_ctx,
            # Use 4mb buffer as sometimes single job response can be huge.
            read_bufsize=2 ** 22,
        ) as resp:
            if 400 <= resp.status:
                err_text = await resp.text()
                if resp.content_type.lower() == "application/json":
                    try:
                        payload = jsonmodule.loads(err_text)
                    except ValueError:
                        # One example would be a HEAD request for application/json
                        payload = {}
                    if "error" in payload:
                        err_text = payload["error"]
                else:
                    payload = {}
                if resp.status == 400 and "errno" in payload:
                    os_errno: Any = payload["errno"]
                    os_errno = errno.__dict__.get(os_errno, os_errno)
                    raise OSError(os_errno, err_text)
                err_cls = self._exception_map.get(resp.status, IllegalArgumentError)
                raise err_cls(err_text)
            else:
                yield resp

    @asyncgeneratorcontextmanager
    async def ws_connect(
        self, abs_url: URL, auth: str, *, headers: Optional[Dict[str, str]] = None
    ) -> AsyncIterator[WSMessage]:
        # TODO: timeout
        assert abs_url.is_absolute(), abs_url
        log.debug("Fetch web socket: %s", abs_url)

        if headers is not None:
            real_headers: CIMultiDict[str] = CIMultiDict(headers)
        else:
            real_headers = CIMultiDict()
        real_headers["Authorization"] = auth

        async with self._session.ws_connect(abs_url, headers=real_headers) as ws:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    yield msg


def _ensure_schema(db: sqlite3.Connection, *, update: bool) -> bool:
    cur = db.cursor()
    ok = True
    found = set()
    cur.execute("SELECT type, name, sql from sqlite_master")
    for type, name, sql in cur:
        if type not in ("table", "index"):
            continue
        if name in SCHEMA:
            if SCHEMA[name] != sql:
                ok = False
                break
            else:
                found.add(name)

    if not ok or found < SCHEMA.keys():
        for sql in reversed(list(DROP.values())):
            cur.execute(sql)
        for sql in SCHEMA.values():
            cur.execute(sql)
        return True
    return False


def _save_cookies(
    db: sqlite3.Connection,
    cookies: Sequence["Morsel[str]"],
    *,
    now: Optional[float] = None,
) -> None:
    if now is None:
        now = time.time()
    _ensure_schema(db, update=True)
    cur = db.cursor()
    for cookie in cookies:
        cur.execute(
            """\
                INSERT OR REPLACE INTO cookie_session
                (name, domain, path, cookie, timestamp)
                VALUES (?, ?, ?, ?, ?)""",
            (cookie.key, cookie["domain"], cookie["path"], cookie.value, now),
        )
    cur.execute(
        "DELETE FROM cookie_session WHERE timestamp < ?", (now - SESSION_COOKIE_MAXAGE,)
    )
    with contextlib.suppress(sqlite3.OperationalError):
        db.commit()


def _load_cookies(
    db: sqlite3.Connection, *, now: Optional[float] = None
) -> List["Morsel[str]"]:
    if now is None:
        now = time.time()
    if _ensure_schema(db, update=False):
        return []
    cur = db.execute(
        """\
            SELECT name, domain, path, cookie, timestamp FROM cookie_session
            WHERE timestamp >= ?
            ORDER BY name
        """,
        (now - SESSION_COOKIE_MAXAGE,),
    )
    ret: List[Morsel[str]] = []
    for name, domain, path, value, timestamp in cur:
        ret.append(_make_cookie(name, value, domain, path))
    return ret


def _make_cookie(name: str, value: str, domain: str, path: str) -> "Morsel[str]":
    tmp = SimpleCookie()  # type: ignore
    tmp[name] = value
    cookie = tmp[name]
    cookie["domain"] = domain
    cookie["path"] = path
    cookie["max-age"] = str(SESSION_COOKIE_MAXAGE)
    return cookie



from ._rewrite import rewrite_module


@rewrite_module
class ConfigError(RuntimeError):
    pass


@rewrite_module
class ClientError(Exception):
    pass


@rewrite_module
class IllegalArgumentError(ValueError):
    pass


@rewrite_module
class AuthError(ClientError):
    pass


@rewrite_module
class AuthenticationError(AuthError):
    pass


@rewrite_module
class AuthorizationError(AuthError):
    pass


@rewrite_module
class ResourceNotFound(ValueError):
    pass


@rewrite_module
class ServerNotAvailable(ValueError):
    pass


@rewrite_module
class ConfigLoadException(Exception):
    pass


@rewrite_module
class NDJSONError(ValueError):
    pass


@rewrite_module
class NotSupportedError(NotImplementedError):
    pass


@rewrite_module
class StdStreamError(Exception):
    def __init__(self, exit_code: int) -> None:
        super().__init__(f"Stream finished with exit code {exit_code}")

        self.exit_code = exit_code
